
"""
    This module is used to run a function in a thread.
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
 
try:
    import threading
except ImportError:
    threading = None
import time
from pyDendron.app_logger import logger, perror

class RunThread:
    def __init__(self, run_action, end_action=None, args=(), kwargs={}):
        self._stop_event = threading.Event()
        self.run_action = run_action
        self.end_action = end_action
        self.kwargs = kwargs 
        self.args = args
        self.kwargs['stop_event'] = self._stop_event
        self.kwargs['end_action'] = self.end_action

    def start(self):
        #perror("Starting thread")
        if threading is None:
            self.run_action(*self.args, **self.kwargs)
            self.end_action()
        else:
            self.thread = threading.Thread(target=self.run_action, args=self.args, kwargs=self.kwargs)
            self.thread.start()
        #perror("Thread started")

    def stop(self):
        #perror("Stopping thread")
        self._stop_event.set()
        self.thread.join()
        #perror("Thread stopped")
        



"""
Modal dialog box
"""
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
__license__ = "GPL"

import panel as pn
import param

from pyDendron.app_logger import logger, perror

class MessageBox:
    def __init__(self, template, title, message, buttons):
        self.template = template
        self.title = title
        self.message = message
        self.buttons = buttons
        self.action = False
        
        self.yes = pn.widgets.Button(name='Yes', button_type='primary', visible=False)
        self.yes.on_click(self.on_yes)
        self.no = pn.widgets.Button(name='No', button_type='primary', visible=False)
        self.no.on_click(self.on_no)
        self.cancel = pn.widgets.Button(name='Cancel', button_type='primary', visible=False)
        self.cancel.on_click(self.on_cancel)
        self.ok = pn.widgets.Button(name='Ok', button_type='primary', visible=False)
        self.ok.on_click(self.on_ok)
        
        self.buttons = pn.Row(self.yes, self.no, self.ok, self.cancel)
        
        for name in buttons:
            if name == 'yes':
                self.yes.visible = True
            elif name == 'no':
                self.no.visible = True
            elif name == 'cancel':
                self.cancel.visible = True
            elif name == 'ok':
                self.ok.visible = True
        
        self.objects =[f'## {self.title}', self.message, self.buttons]
        
    def open(self):
        self.template.modal[0].clear()      
        for obj in self.objects:
            self.template.modal[0].append(obj)  
        self.template.open_modal()
    
    def on_yes(self, event):
        self.action = True
        self.template.close_modal()
        #perror('yes')
    
    def on_no(self, event):
        self.action = False
        self.template.close_modal()
        #perror('no')
    
    def on_cancel(self, event):
        self.action = False
        self.template.close_modal()
        #perror('cancel')
    
    def on_ok(self, event):
        self.action = True
        self.template.close_modal()
        #perror('ok')
    
    # class StringInputBox(MessageBox):
    #     value = Param.String()
    #     def __init__(self, template, title, message):
    #         super().__init__(template, title, message, ['ok', 'cancel'])
    #         self._layout.insert(2, self.param.value)

    # class IntegerInputBox(MessageBox):
    #     value = Param.Integer()
    #     def __init__(self, template, title, message):
    #         super().__init__(template, title, message, ['ok', 'cancel'])
    #         self._layout.insert(2, self.param.value)



"""
Detrend / Indices
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"


import numpy as np
import pandas as pd
from csaps import csaps
#from scipy.interpolate import UnivariateSpline

try:
    from concurrent.futures import ProcessPoolExecutor
except ImportError:
    ProcessPoolExecutor = None

from pyDendron.app_logger import logger, perror
from pyDendron.dataname import *

def worker(detrend_fct, id, vec, ring_type, window_size, do_log):
    return detrend_fct(id, vec, ring_type, window_size, do_log)

def detrend(data, ring_type, window_size=5, do_log=False, num_threads=1):
    if not (data[CATEGORY] == MEASURE).all():
        logger.warning('Detrend : data does not contain only samples')
    if num_threads > 1 and ProcessPoolExecutor is not None:
        #perror(f'detrend: multi-threading {num_threads}')
        return detrend_thread(data, ring_type, window_size, do_log, num_threads)
        
    data_out = data.copy()
    for id, row in data.iterrows():
        _, vector =  detrend_vector(id, row[DATA_VALUES], ring_type=ring_type, window_size=window_size, do_log=do_log)
        data_out.at[id, DATA_VALUES] = vector
        data_out.at[id, DATA_LENGTH] = len(vector)
        data_out.at[id, DATA_TYPE] = ring_type
    
    return data_out

def detrend_thread(data, ring_type, window_size=5, do_log=False, num_threads=1):
    with ProcessPoolExecutor(max_workers=num_threads) as executor:
        futures = []
        # Start an asynchronous task (future) for each index
        for id, row in data.iterrows():
            _window_size = window_size
            if len(row[DATA_VALUES]) < window_size:
                logger.warning(f'detrend_vector: in {row[id, KEYCODE],} data length < window size ({len(row[DATA_VALUES])} < {window_size})')
                _window_size = len(row[DATA_VALUES])
            future = executor.submit(worker, detrend, id, row[DATA_VALUES], ring_type, _window_size, do_log)
            futures.append(future)

        # Wait for all tasks to be completed
        data_out = data.copy()
        for future in futures:
            id, vector = future.result()
            data_out.at[id, DATA_VALUES] = vector
            data_out.at[id, DATA_LENGTH] = len(vector)
            data_out.at[id, DATA_TYPE] = ring_type
        return data_out

def detrend_vector(id, vec, ring_type, window_size=5, do_log=False):
    if ring_type == BESANCON:
        return id, besancon(vec, window_size=window_size, do_log=do_log, after=False)
    if ring_type == BESANCON1:
        return id, besancon(vec, window_size=window_size, do_log=do_log, after=True)
    elif ring_type == HANNING:
        return id, sliding_window(vec, window_type=HANNING, window_size=window_size, do_log=do_log)
    elif ring_type == HAMMING:
        return id, sliding_window(vec, window_type=HAMMING, window_size=window_size, do_log=do_log)
    elif ring_type == BARTLETT:
        return id, sliding_window(vec, window_type=BARTLETT, window_size=window_size, do_log=do_log)
    elif ring_type == BLACKMAN:
        return id, sliding_window(vec, window_type=BLACKMAN, window_size=window_size, do_log=do_log)
    elif ring_type == RECTANGULAR:
        return id, sliding_window(vec, window_type=RECTANGULAR, window_size=window_size, do_log=do_log)
    elif ring_type == BP73:
        return id, sliding_window(vec, window_type=RECTANGULAR, window_size=window_size, do_log=True)
    elif ring_type == SPLINE:
        return id, spline(vec, do_log=do_log)
    elif ring_type == SLOPE:
        return id, slope(vec)
    elif ring_type == CORRIDOR:
        return id, corridor(vec, do_log=do_log)
    elif ring_type == CORRIDOR_SPLINE:
        return id, corridor_spline(vec, do_log=do_log)
    elif ring_type == DELTA:
        return id, delta(vec, 1)
    elif ring_type == DELTADELTA:
        return id, delta(vec, 2)
    elif ring_type == LOG:
        return id, np.log(vec)
    elif ring_type == RAW:
        return id, vec
    else:
        raise ValueError(f'detrend_vector: unknown method: {ring_type}')

def _spline(x, raw_vector):
    amp = 0.5
    period = len(raw_vector) * 0.67 #[p114 Cook, 1990]
    freq = 1/period
    # ~ eq 3.13 [Cook, 1990]
    param = 1/(((np.cos(2 * np.pi * freq) + 2) * (1 - amp)/(12 * amp * (np.cos(2 * np.pi * freq) - 1) ** 2))+ 1)
    mask = ~np.isnan(raw_vector)
    return csaps(np.array(x)[mask], raw_vector[mask], x, smooth=param)

# def _spline(x, raw_vector):
#     amp = 0.5
#     period = len(raw_vector) * 0.67  # [p114 Cook, 1990]
#     freq = 1 / period
#     param = 1 / (((np.cos(2 * np.pi * freq) + 2) * (1 - amp) / (12 * amp * (np.cos(2 * np.pi * freq) - 1) ** 2)) + 1)
#     mask = ~np.isnan(raw_vector)
#     spline = UnivariateSpline(np.array(x)[mask], raw_vector[mask], s=param)
#     return spline(x)

def spline(raw_vector, do_log=False):
    x = [x for x in range(len(raw_vector))]
    vector = raw_vector / _spline(x, raw_vector)
    if do_log:
        vector[vector == 0] += np.finfo(float).eps
        vector = np.log(vector)
    return np.round(100*vector, 3)

# def spline_dpl(raw_vector):
#     import dplPy
    
#     df = pd.DataFrame(raw_vector, columns=['data'])    
#     return np.round(dplPy.detrend(df, fit="spline", plot=False)['data'].to_numpy()*100, 3)

def vector_log(vector, copy=True):
    if copy:
        vector = vector.copy()
    vector[vector == 0] += np.nan
    mask = ~np.isnan(vector)
    vector[mask] = np.log(vector[mask])
    return vector
    

def _sliding_window(raw_vector, window):
    vector = np.copy(raw_vector)
    n = len(window)
    view = np.lib.stride_tricks.sliding_window_view(vector, window_shape=n).copy()
    
    view *= window
    conv = np.sum(view, axis=1)
    tmp = np.full(n // 2, np.nan)
    conv = np.concatenate((tmp, conv, tmp))
    vector /= conv
    vector *= 100 # Scale vector
    
    return vector

def sliding_window(raw_vector, window_type=HANNING, window_size: int = 5, do_log=False):    
    if window_type == HANNING:
        window = np.hanning(window_size)
    elif window_type == HAMMING:
        window = np.hamming(window_size)
    elif window_type == BARTLETT:
        window = np.bartlett(window_size)
    elif window_type == BLACKMAN:
        window = np.blackman(window_size)
    elif window_type == RECTANGULAR:
        window = np.ones(window_size)
    else:
        raise ValueError('sliding_window: unknown window_type: {window_type}')
    window /= window.sum()
    
    vector = _sliding_window(raw_vector, window)
    if do_log:
        vector = vector_log(vector, copy=False)
    
    #print('VERIFIER LA DIMENSSION DES VECTORS IN OUT !!!' , len(raw_vector), len(vector))

    return np.round(vector, 3)

def besancon(raw_vector, window_size, do_log=True, after=False):
    """
    Calculate the Besancon indice for the given DataFrame.

    :param window_size: The size of the sliding window for indice calculation. Default is 7.
    :type window_size: int
    
    :param log_before: Compute the log transformation befor any transformation. 
    :type log_before: bool
    """
    n = window_size // 2
    # Create a view using as_strided to obtain the sliding window
    vector = vector_log(raw_vector) if do_log and (not after) else raw_vector
    view = np.lib.stride_tricks.sliding_window_view(vector, window_shape=window_size).copy()
    vector = view[:, n+1].copy()
    # Set the central value to nan
    view[:, n+1] = np.nan
    vector *= (np.sum(~np.isnan(view), axis=1) - 2) * 100
    denominator = (np.nansum(view, axis=1) - np.nanmax(view, axis=1) - np.nanmin(view, axis=1))
    denominator[denominator == 0] = np.nan
    mask = np.isnan(vector) | np.isnan(denominator)
    vector[mask] = np.nan
    vector[~mask] /= denominator[~mask]
    if do_log and after:
        vector = vector_log(vector, copy=False)
    tmp = np.full(n, np.nan)
    vector = np.concatenate((tmp, np.round(vector, 3), tmp))
    return vector

def slope(raw_vector, slope_resolution=0):
    """
    Calculate the slope indice ('gleichlaufigkeit') .
    """
    # Calculate the first-order discrete difference
    diff_series = np.diff(raw_vector)
    if slope_resolution >= 0:
        diff_series[np.abs(diff_series) < slope_resolution] = 0
    # Compute the sign (-1 if x < 0, 0 if x==0, 1 if x > 0) of each element 
    sign_series = np.sign(diff_series)
    # Append a NaN value to the end of 'sign_series' to match the original series size
    vector = np.append(np.nan, sign_series)
    return vector

def signatures(data_col, slope_resolution=0, min_elements=4):
    """
    Calculate the signatures 
    """
    
    df = data_col.apply(lambda col: slope(col.to_numpy(), slope_resolution), axis=0)
    mask = (df.shape[1] - df.isnull().sum(axis=1)) < min_elements
    
    df = df.mean(axis=1).abs()
    df[mask] = np.nan

    #perror(f'signatures: {df.shape}')
    #perror(f'signatures mean: {df.to_numpy()}')

    return df.to_numpy()
    
def delta(raw_vector, order=1):
    """
    Calculate the derivate indice .
    """
    #vector = np.copy(raw_vector)
    # Calculate the first-order discrete difference
    vector = np.diff(raw_vector, n=order, prepend=[np.nan]*order)
    #for i in range(order):
    #    diff_series = np.diff(vector)
    #    vector = np.append(diff_series, np.nan)

    return vector

def corridor(raw_vector, do_log=False):
    # 
    vector = np.copy(raw_vector) # copy of the original vector not needed ?
    mu = np.nanmean(vector) # mean without NaN value
    std = np.nanstd(vector) # std without NaN value
    thr = mu + 2 * std # Threshold 

    mask_thr = vector > thr # binary mask, True if value > threshold

    windows = np.lib.stride_tricks.sliding_window_view(mask_thr, window_shape=3)
    mask_filters = np.concatenate(( [False], np.sum(windows == [False, True, False], axis=1) == 3, [False]))
    value_filters = np.concatenate(( [np.nan], np.convolve(vector, [1,0,1], mode='valid')/2, [np.nan]))
    vector[mask_filters] = value_filters[mask_filters]

    mask_thr[mask_filters] = False
    vector[mask_thr] = np.nan

    nan_mask = ~np.isnan(vector)

    x = np.array([x for x in range(len(vector))])
    x_masked = x[nan_mask]
    vec_masked = vector[nan_mask]

    # Polynomial regression based on corrected values
    coef_middle = np.polyfit(x_masked, vec_masked, 3)
    fct_middle = np.poly1d(coef_middle)
    #vec_middle = fct_middle(x)
    vec_middle_masked = fct_middle(x_masked)

    # Polynomial regression based on roof values
    mask = (vec_masked - vec_middle_masked) >= 0
    coef_roof = np.polyfit(x_masked[mask], vec_masked[mask], 3)
    fct_roof = np.poly1d(coef_roof)
    vec_roof = fct_roof(x)

    # Polynomial regression based on floor values
    mask = (vec_masked - vec_middle_masked) < 0
    coef_floor = np.polyfit(x_masked[mask], vec_masked[mask], 3)
    fct_floor = np.poly1d(coef_floor)
    vec_floor= fct_floor(x)

    # detrended values
    vector = (vector - vec_floor) / (vec_roof - vec_floor)
    min_ = np.nanmin(vector)
    max_ = np.nanmax(vector)
    min_raw = np.nanmin(raw_vector)
    max_raw = np.nanmax(raw_vector)

    vector = np.round((vector - min_ ) * (max_raw - min_raw) / (max_ - min_) + min_raw, 2)
    
    if do_log:
        vector = vector_log(vector, copy=False)
   
    return vector

def corridor_spline(raw_vector, do_log):
    # 
    vector = np.copy(raw_vector) # copy of the original vector not needed ?
    mu = np.nanmean(vector) # mean without NaN value
    std = np.nanstd(vector) # std without NaN value
    thr = mu + 2 * std # Threshold 

    mask_thr = vector > thr # binary mask, True if value > threshold
    vector[mask_thr] = np.nan
    #raw_vector[mask_thr] = np.nan

    nan_mask = ~np.isnan(vector)

    x = np.array([x for x in range(len(vector))])

    # Spline based on corrected values
    vec_middle = _spline(x, vector)
    
    mask = (vector - vec_middle) > 0
    vec_roof = _spline(x, vector[mask])
    
    mask = (vector - vec_middle) < 0
    vec_floor = _spline(x, vector[mask])

    # detrended values
    vector = np.copy(raw_vector) # copy of the original vector not needed ?
    vector = (vector - vec_floor) / (vec_roof - vec_floor) 
    
    min_ = np.nanmin(vector)
    max_ = np.nanmax(vector)
    min_raw = np.nanmin(raw_vector)
    max_raw = np.nanmax(raw_vector)

    vector = np.round((vector - min_ ) * (max_raw - min_raw) / (max_ - min_) + min_raw, 2)

    if do_log:
        vector = vector_log(vector, copy=False)
    
    return vector




"""
Dataset class
"""
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
__license__ = "GPL"

from typing import List, Tuple, Union
import warnings
import os
import copy
import pickle
import json
from pathlib import Path
from datetime import datetime
from concurrent.futures import ProcessPoolExecutor

import pandas as pd
import numpy as np
import panel as pn
import param

from collections import Counter

from pyDendron.app_logger import logger, perror
from pyDendron.dataname import *
from pyDendron.componentsTree import ComponentsNode, ComponentsTree
from pyDendron.mean import compute_mean, compute_means
from pyDendron.crossdating import CrossDating
from pyDendron.detrend import detrend, slope
from pyDendron.estimation import cambium_estimation

class Dataset(param.Parameterized):
    """
    Data storage of sequences and components. Includes also selections of pairs.
    """
    VERSION = 11
    
    notify_message = param.String(default='', doc='log change in the dataset') 
    notify_reload = param.Event()
    notify_synchronize = param.Event()
    notify_packages = param.Event()
    notification_source = param.ClassSelector(default=None, class_=object)
    #counter = param.Integer(3, doc='Node added in tree')
    save_auto =  param.Boolean(False, doc='show all components / sequences')
            
    version_control = True
    
    def __init__(self, sequences=None, components=None, username='None', cfg_tmp='./', **params):
        super(Dataset, self).__init__(**params)   
        self.username = username        
        self.filename = None
        self.cfg_tmp = cfg_tmp
        
        if (components is not None) and (sequences is not None):
            self.sequences = pd.DataFrame(sequences)
            self.components = pd.DataFrame(components)
            self.update()
        else:
            self.clean()   
        self._packages = {}
        self._freeze_components = None
        self._freeze_sequences = None
        self._log = []
        self._crossdating = pd.DataFrame()
    
    def get_log(self):
            """
            Returns the log data as a pandas DataFrame.

            Returns:
                pandas.DataFrame: The log data.
            """
            #perror('get_log', self._log)
            return pd.DataFrame(self._log, columns=log_dtype.keys())
    
    def get_sequences_copy(self, ids):
        """
        Returns a copy of the sequences at the specified indices.

        Parameters:
            ids (list): A list of indices specifying the sequences to be copied.

        Returns:
            pandas.DataFrame: A copy of the sequences at the specified indices.
        """
        data = self.sequences.loc[ids,:].copy()
        return data

    def get_components_copy(self, id_pairs):
        """
        Returns a copy of the components DataFrame based on the given index pairs.

        Parameters:
            id_pairs (list): A list of index pairs specifying the rows to be copied.

        Returns:
            pandas.DataFrame: A copy of the components DataFrame containing the specified rows.
        """
        data = self.components.loc[id_pairs,:].copy()
        return data
        
    def freeze_sequences(self, ids):
        """
        Freezes the sequences at the specified indices.

        Args:
            ids (list): A list of indices indicating the sequences to freeze.

        Returns:
            None
        """
        #perror('freeze_sequences', ids)
        self._freeze_sequences = self.get_sequences_copy(ids)

    def freeze_components(self, id_pairs):
        """
        Freezes the components specified by the given index pairs.

        Args:
            id_pairs (list): A list of index pairs specifying the components to freeze.

        Returns:
            None
        """
        self._freeze_components = self.get_components_copy(id_pairs)

    def log_components(self, id_pairs, comments=''):
        """
        Compare the components specified by the given index pairs.

        Args:
            id_pairs (list): A list of index pairs specifying the components to compare.

        Returns:
            None

        Raises:
            None
        """
        new_df = self.get_components_copy(id_pairs)
        old_df = self._freeze_components
        log = []
        
        merge = old_df.join(new_df, lsuffix='_old', rsuffix='_new')
        for ids, row in merge.iterrows():
            (id_child, id_parent) = ids
            old, new = row[OFFSET+'_old'], row[OFFSET+'_new']
            if pd.isna(old) or pd.isna(new) or (old != new):
                log.append([datetime.now(), id_child, id_parent ,OFFSET, old, new, self.username, comments])
        
        self._freeze_components = None
        if len(log) > 0:
            self._log += log
            #self.notify_changes(comments)
            return True
        #print('no change in components')
        return False
        
    def log_sequences(self, ids, comments=''):
        """
        Compare sequences between the old and new dataframes.

        Args:
            ids (list): List of indices to compare.
            comments (str, optional): Additional comments for the comparison. Defaults to ''.

        Returns:
            dict: A dictionary containing the history of changes made in the sequences.
                The dictionary keys are tuples of (index, column, date), and the values are lists
                containing the old value, new value, user, and comments.

        Raises:
            KeyError: If the two dataframes are not aligned.

        """
        def compare(id, old_row, new_row):
            log = []
            #d = new_row[DATE_SAMPLING]
            d = datetime.now()
            for col in new_row.index:
                differ = False 
                old, new = old_row[col], new_row[col]
                #perror(f'log_sequences compare {id}, {col}, {type(old)}, {type(new)}, ||{old}||, ||{new}||')
                if isinstance(old, (np.ndarray, pd.Series)) and isinstance(new, (np.ndarray, pd.Series)):
                    differ = not np.array_equal(np.nan_to_num(old), np.nan_to_num(new))
                elif (col == DATA_INFO) and isinstance(old, list) and isinstance(new, list):
                    differ = sorted(old) == sorted(new)
                elif pd.isna(old) and pd.isna(new):
                    differ = False
                elif pd.isna(old) or pd.isna(new):
                    differ = True
                else:
                    differ = old != new
                if differ:
                    log.append([d, id, pd.NA ,col, old, new, self.username, comments])
            return log
        
        old_df = self._freeze_sequences
        new_df = self.get_sequences_copy(ids)
        log = []
        if isinstance(new_df, pd.Series):
            log = compare(ids, old_df, new_df)
        else:
            for (id1, row1), (id2, row2) in zip(old_df.iterrows(), new_df.iterrows()):
                if id1 == id2:
                    log = compare(id1, row1, row2)
                else:
                    raise KeyError('The 2 dataframes are not aligned.')
        self._freeze_sequences = None
        if len(log) > 0:
            self._log += log
            return True
        return False
    
    def get_crossdating_log(self):
        """
        Returns the crossdating log associated with the dataset.

        Returns:
            The crossdating log.
        """
        return self._crossdating
    
    def log_crossdating(self, crossdating):
        """
        Logs the crossdating information for a dataset.

        Parameters:
        crossdating (dict): A dictionary containing the crossdating information.

        Returns:
        None
        """
        crossdating[CROSSDATING_DATE] = datetime.now()
        if len(self._crossdating) > 0:
            self._crossdating.loc[self._crossdating.index.max()+1] = crossdating
        else:
            self._crossdating = pd.DataFrame([crossdating])
        
    
    def is_empty(self):
        """
        Returns True if dataset is empty.
        """
        return len(self.sequences) == 0
        
    def notify_changes(self, message, source=None):
        """
        Set `msg` into `self.change`.
        """
        self.notify_message = message
        if self.save_auto:
            #print('*** Save auto ***')
            self.dump()
            
        self.notification_source = source if source is not None else self
        if message in ['load', 'reindex']:
            self.param.trigger('notify_reload')
        else:
            self.param.trigger('notify_synchronize')
        
    def update(self, update_tree=False):
        """
        Update index and dtype of `self.sequences` and `self.component`.
        """
        self.sequences.reset_index(inplace=True)
        self.sequences = pd.DataFrame(self.sequences, columns=sequences_index+sequences_cols)
        self.sequences.set_index(sequences_index, inplace=True, verify_integrity=True)  
        self.sequences = self.sequences.astype(sequences_dtype_dict, copy=True, errors='ignore')
        self.components.reset_index(inplace=True)
        self.components = pd.DataFrame(self.components, columns=components_index+components_cols)
        self.components.set_index(components_index, inplace=True, verify_integrity=True)  
    
    def clone(self):
        """
        Copy the dataset and returns it.
        """
        tmp = Dataset()
        tmp.sequences = self.sequences.copy()
        tmp.components = self.components.copy()
        tmp._packages = copy.deepcopy(self._packages)
        
        return tmp
    
    def _get_filename(self, filename: str = None) -> str:
        if filename is None:
            if self.filename is None:
                raise ValueError('DataSet._get_filename: empty filename')
            else:
                filename = self.filename
        else:
            self.filename = filename
        return filename

    def backup(self, filename: str):
        """
        Dump/save a dataset into `filename`.
        """
        filename = Path(filename)
        current_datetime = datetime.now().strftime("%Y%m%d_%H%M%S")
        new_filename = f"{filename.stem}_{current_datetime}{filename.suffix}"
        #print('backup', filename)
        # Construct the new full file path
        directory = filename.parent 
        new_file_path = directory / 'backup' / new_filename
        os.makedirs(directory / 'backup', exist_ok=True)

        self.dump(new_file_path, save_name=False)
 
    def dump(self, filename: str = None, save_name=True):
        """
        Dump/save a dataset into `filename`.
        """
        filename = self._get_filename(filename) if save_name else filename
        #print('save', filename, self.filename)
        suffix = Path(filename).suffix
        if suffix == '.json':
            self._dump_json(filename)
        elif suffix == '.p':
            self._dump_pickle(filename)
        elif suffix == '.xlsx':
            self._dump_excel(filename)
        else:
            raise TypeError(f'DataSet.dump: unknown suffix {suffix} from {filename}')
        #perror('dump', self.components.index)
        #perror('dump', self.components.index)

    def _dump_pickle(self, filename: str):
        dataset_path = Path(filename) 
        with dataset_path.open('wb') as fic:
            #perror('_dump_pickle', self._log)
            pickle.dump((self.VERSION, self.sequences, self.components, self._packages, self._log, self._crossdating), fic)

    def _dump_json(self, filename: str):
        dataset_path = Path(filename) 
        with dataset_path.open('w') as fic:
            dfs_json = {
                VERSION: self.VERSION,
                SEQUENCES: json.loads(self.sequences.to_json(orient='table', index=True, force_ascii=False, indent=2)),
                COMPONENTS: json.loads(self.components.to_json(orient='table', index=True, force_ascii=False, indent=2)),
                PACKAGES: self._packages,
                LOG: json.loads(self.get_log().to_json(orient='table', index=True, force_ascii=False, indent=2)),
                CROSSDATING: json.loads(self._crossdating.to_json(orient='table', index=True, force_ascii=False, indent=2))
            }
            json.dump(dfs_json, fic, indent=2)            

    def _dump_excel(self, filename: str):
        filename = Path(filename) 
        with pd.ExcelWriter(filename) as writer:
            self.sequences.to_excel(writer, sheet_name=SEQUENCES, merge_cells=False, float_format="%.6f")
            self.components.to_excel(writer, sheet_name=COMPONENTS, merge_cells=False, float_format="%.6f")

    def _dump_csv(self, path: str):
        base_path = Path(path)
        self.sequences.to_csv(base_path / 'sequences.csv', sep='\t', float_format="%.6f")
        self.components.to_csv(base_path / 'components.csv', sep='\t', float_format="%.6f")

    def update_version(self, version):
        """
        Update the dataset to a specific version.

        Parameters:
        - version (int): The version number to update the dataset to.

        Returns:
        - None

        Raises:
        - None
        """
            
        if version == self.VERSION:
            return
        
        self._log = []
        self._crossdating = pd.DataFrame()
        
        if version <= 1:
            self.sequences[COMPONENT_COUNT] = pd.NA
        if version <= 2:
            mask = self.sequences[CATEGORY] == 'Tree'
            self.sequences.loc[mask, CATEGORY] = MEASURE
            mask = self.sequences[CATEGORY] == 'Chronology'
            self.sequences.loc[mask, CATEGORY] = MEAN
        if version <= 3:
            self.sequences[MEANAS_MEASURE] = False
        if version <= 4:
            if 'Dated' in self.sequences.columns:
                self.sequences.drop(columns=['Dated'], inplace=True)
        if version <= 5:
            if 'tag' in self.sequences.columns:
                self.sequences[TAG] = self.sequences['tag']
                self.sequences.drop(columns=['tag'], inplace=True)
        if version <= 6:
            if 'PithAbsent' in self.sequences.columns:
                self.sequences.drop(columns=['PithAbsent'], inplace=True)
        if version <= 7:
            self.sequences.index.name = sequences_index[0]
            self.components.index.names = components_index  
        if version <= 11:
            for index, row in self.sequences.iterrows():
                length = row[DATA_LENGTH]  # Taille du tableau de zéros
                if pd.notna(length) and length > 0:
                    self.sequences.at[index, DATA_SIGNATURES] = np.zeros(length)
                else:
                    self.sequences.at[index, DATA_SIGNATURES] = pd.NA
        
        logger.info(f'update dataset version {version} to {self.VERSION}')

    def load(self, filename: str=None):
        """
        Load a dataset from `filename`.
        """
        filename = self._get_filename(filename)
        ('load', filename)
        suffix = Path(filename).suffix
        version = self.VERSION
        if suffix == '.json':
            version = self._load_json(filename)
        elif suffix == '.p':
            version = self._load_pickle(filename)
        elif suffix == '.xlsx':
            self._load_excel(filename)
        else:
            raise TypeError(f'DataSet.load: unknown suffix {suffix} from {filename}')
        self.update_version(version)
        self.notify_changes('load')
        #perror('roots', self.get_roots())

    def _load_pickle(self, filename: str):
        dataset_path = Path(filename) 
        with dataset_path.open('rb') as fic:
            data = pickle.load(fic)
            version, self.sequences, self.components, self._packages, self._log, self._crossdating = data
            #perror('load', self._log)
        return version

    def _load_json(self, filename):
        def json_to_df(json_data, key):
            data = json_data[key]["data"]
            
            # Extraire le schéma des colonnes
            schema_fields = json_data[key]["schema"]["fields"]
            columns = [field["name"] for field in schema_fields]

            # Créer le DataFrame avec les colonnes et les données
            return pd.DataFrame(data, columns=columns)
      
       
        dataset_path = Path(filename) 
        with dataset_path.open('r') as fic:
            dfs_json = json.load(fic)
        version = dfs_json[VERSION]
        self.sequences = json_to_df(dfs_json, SEQUENCES)
        self.sequences[DATA_VALUES] = self.sequences[DATA_VALUES].apply(lambda x: np.array(x)) 
        self.sequences[DATA_WEIGHTS] = self.sequences[DATA_WEIGHTS].apply(lambda x: np.array(x))
        self.sequences[DATA_SIGNATURES] = self.sequences[DATA_SIGNATURES].apply(lambda x: np.array(x))
        
        self.components = json_to_df(dfs_json, COMPONENTS)    
        self._packages = dfs_json[PACKAGES]
        self._log = json_to_df(dfs_json, LOG)
        self._crossdating = json_to_df(dfs_json, CROSSDATING)
        self.update()
        return version
    
    def _load_excel(self, filename: str):
        filename = Path(filename) 
        seqs = pd.read_excel(filename, sheet_name=SEQUENCES)
        comps = pd.read_excel(filename, sheet_name=COMPONENTS)
        self.sequences = pd.DataFrame(seqs)
        self.components = pd.DataFrame(comps)        
        self.update()

    def _load_csv(self, path: str):
        base_path = Path(path)
        seqs = pd.read_csv(base_path / 'sequences.csv', sep='\t')
        comps = pd.read_csv(base_path / 'components.csv', sep='\t')
        self.sequences = pd.DataFrame(seqs)
        self.components = pd.DataFrame(comps)        
        self.update()
        
    def new_dataset(cls):
        """
        Create a new dataset.

        Returns:
            dataset (cls.Dataset): The newly created dataset.
        """
        dataset = cls.Dataset()
        dataset.new_root()
        dataset.new_trash()
        dataset.new_workshop()
        return dataset
    
    def new_root_id(self):
        """
        Create a new root node in the dataset.

        Returns:
            int: The index of the new root node.
        """
        return self.sequences.index.min() - 10
    
    def new_root(self, keycode: str = 'Dataset', id = ROOT, append=True):
        """
        Create a new root node in the dataset.

        Args:
            keycode (str, optional): The keycode for the new root node. Defaults to 'Dataset'.
            id (int, optional): The index of the new root node. Defaults to ROOT.

        Returns:
            int: The index of the new root node.

        """
        self.new(keycode, SET, id_parent=None, id=id)
        if append:
            data = []
            for id_child in set(self.get_roots()):
                if id != id_child:
                    data.append({ID_PARENT: id, ID_CHILD: id_child, OFFSET: pd.NA})
            df = pd.DataFrame(data).set_index(components_index, verify_integrity=True)
            with warnings.catch_warnings():
                # TODO: pandas 2.1.0 has a FutureWarning for concatenating DataFrames with Null entries
                warnings.filterwarnings("ignore", category=FutureWarning)
                self.components = pd.concat([self.components, df])
        return id
            
    def new_trash(self, keycode: str = 'Trash'):
        """
        Create a new trash item in the dataset.

        Parameters:
            keycode (str): The keycode for the trash item. Defaults to 'Trash'.

        Returns:
            The newly created trash item.
        """
        return self.new(id=TRASH, id_parent=None, keycode=keycode, category=SET)

    def new_workshop(self, keycode: str = 'Workshop' ):
        """
        Create a new workshop in the dataset.

        Parameters:
            keycode (str): The keycode for the workshop. Defaults to 'Workshop'.

        Returns:
            The newly created workshop.
        """
        return self.new(id=WORKSHOP, id_parent=None, keycode=keycode, category=SET)

    def new_clipboard(self, keycode: str = 'Clipboard' ):
        """
        Create a new clipboard entry in the dataset.

        Parameters:
            keycode (str): The keycode for the clipboard entry. Defaults to 'Clipboard'.

        Returns:
            The newly created clipboard entry.
        """
        return self.new(id=CLIPBOARD, id_parent=None, keycode=keycode, category=SET)

    def new(self, keycode: str, category: str, id_parent: int | None, 
            id: int | None = None, others = {}, offset : int = pd.NA) -> int:
        """
        Creat a new Sequence and component if id_parent is not None.
        
        Arguments
        ---------
        keycode: KEYCODE of the new Sequence.
        category: CATEGORY of the new Sequence.
        id_parent: ID_PARENT of the new Conponent.
        id: ID of the Sequence.
        others: dictionary of field, value pairs to set in new Sequence.
        offset: offset to set in new Component.
        make_root : !!!! error !!! make_root and id_parent
        
        
        Returns
        -------
        The ID of the new Sequence.
        """
        id = self.sequences.index.max() + 1 if id is None else id
        others.update({KEYCODE: keycode, CATEGORY: category})
        if CREATION_DATE not in others:
            others[CREATION_DATE] = datetime.now()
        self.sequences.loc[id, list(others.keys())] = others
        if id_parent is not None:
            self.components.loc[(id_parent, id), OFFSET] = offset
        #self.notify_changes(f'new')
        return id                

    def _id_list(self, ids):
        """
        Returns a list of int created form a List[int] of from int.
        """
        if not isinstance(ids, list):
            ids = [ids]
        return ids
 
    def copy(self, triplets: List[Tuple[int, int, int]], dest_path:  List[int], notify=True) -> str:
        """
        Copy `triplets` in dest_path.
        
        Arguments
        ---------
        triplets: list of tuples (ID_PARENT, ID_CHILD, OFFSET)
        dest_path: a path
        
        Returns
        -------
        A dictonary of destination data : {(ID_PARENT, ID_CHILD): OFFSET}
        """
        # detect circular referencies
        couples_ = [(p, c) for p, c, o in triplets if dest_path[-1] != p]
        if len(couples_) != len(triplets):
            logger.warning('Destination and source are equal. Copy aborded.')
            return None

        couples_ = [(p, c) for p, c, o in triplets if c not in dest_path] 
        if len(couples_) != len(triplets):
            logger.warning('circular reference. Copy aborded.')
            return None

        dest_map = {(dest_path[-1], c) : o for p, c, o in triplets}
        msg = ''
        for keys, offset in dest_map.items():
            if keys in self.components.index:
                msg= str(keys[1])+', '
            else :
                self.components.loc[keys, OFFSET] = offset
        if msg != '':
            msg = 'Duplicates not copied: ' + msg 
            logger.warning(msg)

        if notify:
            self.notify_changes('copy')
                            
        return dest_map
    
    def move(self, triplets: List[Tuple[int, int, int]], dest_path:  List[int], notify=True) -> str:
        """
        Cut `triplets` in dest_path.
        
        Arguments
        ---------
        triplets: list of tuples (ID_PARENT, ID_CHILD, OFFSET)
        dest_path: a path
        
        Returns
        -------
        A dictonary of destination data : {(ID_PARENT, ID_CHILD): OFFSET}
        """
        dest_map = self.copy(triplets, dest_path)
        if dest_map is not None:
            keys = [(p, c) for p, c, o in triplets if p != -1]
            #logger.info(f'dataset cut keys: {keys}')
            self.components.drop(index=keys, inplace=True)
            if notify:
                self.notify_changes('cut')
        return dest_map
        
    def drop(self, triplets: List[Tuple[int, int, int]], notify=True)-> str:
        """
        drop `triplets`.
        
        Arguments
        ---------
        triplets: list of tuples (ID_PARENT, ID_CHILD, OFFSET)
        
        Returns
        -------
        List of sequences droped
        """
        def drop_children(node, cpt_drops, drop_comp, drop_seq, deep=0):
            if cpt_drops[node.id] >= 0:
                #perror('cpt_drops[node.id]', cpt_drops[node.id])
                drop_comp.append((node.parent.id, node.id))
                drop_seq.append(node.id)
                for child in node.children:
                    drop_children(child, cpt_drops, drop_comp, drop_seq, deep+1)
            else:
                if deep == 0:
                    drop_comp.append((node.parent.id, node.id))
                logger.warning(f'Child {node.id} / {node.keycode} has {cpt_drops[node.id]} is dupplicated.')
        
        id_roots = self.get_roots()
        tree_roots = self.get_descendants(id_roots)
        cpt_roots = tree_roots.count_descendants()
        #perror('cpt_roots', cpt_roots)
        
        id_drops = list(set([c for p, c, o in triplets]))
        tree_drops = self.get_descendants(id_drops)
        cpt_drops = tree_drops.count_descendants()
        
        #perror('cpt_drops', cpt_drops)
        
        for key in cpt_drops.keys():
            if key in cpt_roots.keys():
                #perror('cpt_drops[key]', key, cpt_drops[key])
                #perror('cpt_roots[key]', key, cpt_roots[key])
                cpt_drops[key] = cpt_drops[key] - cpt_roots[key]

        #perror('cpt_drops', cpt_drops)
        
        drop_seq = []
        drop_comp = []
        
        for child in tree_drops.children:
            drop_children(child, cpt_drops, drop_comp, drop_seq)

    def soft_drop(self, pairs: List[Tuple[int, int]]) -> str:
        """
        soft drop of `triplets` in trash set.
        
        Arguments
        ---------
        pairs: list of tuples (ID_PARENT, ID_CHILD)

        Returns
        -------
        A string with duplicate sequences erased in trash.
        """
        return self.move(pairs, dest_path=[TRASH])
        
    def clean(self):
        """
        Remove data in `self.sequences` and `self.components` 
        """
        self.sequences = pd.DataFrame()
        self.components = pd.DataFrame()
        #self.notify_changes(f'clean')

    def append(self, dataset, verify_integrity=True, notify=True, merge_root=False):
        """
        Append a dataset to `self`. Warning use pd.concat with NA values.
        deprecated: use `merge` instead.
        """
        if len(self.sequences) > 0:
            tmp = dataset.clone()
            tmp.sequences.drop(index=WORKSHOP, inplace=True)
            tmp.sequences.drop(index=CLIPBOARD, inplace=True)
            tmp.sequences.drop(index=TRASH, inplace=True)
            roots = tmp.get_roots()
            #perror('get roots before drop', roots)
            if (tmp.components.index.get_level_values(ID_PARENT) == WORKSHOP).sum() > 0:
                tmp.components.drop(index=WORKSHOP, inplace=True)
            if (tmp.components.index.get_level_values(ID_PARENT) == CLIPBOARD).sum() > 0:
                tmp.components.drop(index=CLIPBOARD, inplace=True)
            if (tmp.components.index.get_level_values(ID_PARENT) == TRASH).sum() > 0:
                tmp.components.drop(index=TRASH, inplace=True)
            min_id = self.sequences.index.min() - 10
            roots = tmp.get_roots()
            #perror('get roots after drop', roots)
            index_mapping = dict(zip(roots, [i*-10+min_id for i in range(len(roots))]))
            #print(index_mapping)
            tmp.components.rename(index=index_mapping, level=ID_PARENT, inplace=True)
            tmp.sequences.rename(index=index_mapping, level=ID, inplace=True)
            
            tmp.reindex(start=self.sequences.index.max() + 1)
            with warnings.catch_warnings():
                # TODO: pandas 2.1.0 has a FutureWarning for concatenating DataFrames with Null entries
                warnings.filterwarnings("ignore", category=FutureWarning)
                #print(keep)
                self.sequences = pd.concat([self.sequences, tmp.sequences], verify_integrity=verify_integrity)
                self.components = pd.concat([self.components, tmp.components], verify_integrity=verify_integrity)
        else:
            self.sequences = dataset.sequences.copy()
            self.components = dataset.components.copy()
        if notify:
            self.notify_changes('append')
              
    def reindex(self, start=0, notify=True) -> int:
        """
        Reindex sequences from `start` to `start` + number of sequences. 
        Modifies ID_CHILD and IXD_PARENT values in components.

        deprecated: never use.
        
        Returns
        -------
        the last ID        
        """
        # Reindexing with contiguous index
        last = start + len(self.sequences.loc[self.sequences.index >= 0])
        new_index = list(range(start, last))
        # Create a mapping dictionary between old and new index
        index_mapping = dict(zip(self.sequences.index[self.sequences.index >= 0], new_index))
        # Use the dictionary to reindex
        tmp = self.components.rename(index=index_mapping, level=ID_PARENT)
        self.components = tmp.rename(index=index_mapping, level=ID_CHILD)
        self.sequences = self.sequences.rename(index=index_mapping, level=ID)
        if notify:
            self.notify_changes('reindex')
        return last

    def drop_orphans_components(self, paires, level=ID_PARENT):
        self.components = self.components.drop(paires)

    def get_orphans_sequences(self):
        # get sequences that are not TRASH, WORKSHOP, CLIPBOARD, ...
        # root nodes should be < 0
        seqs = set(self.sequences.index[self.sequences.index > 0].tolist())
        parent_ids = set(self.components.index.get_level_values(ID_PARENT).tolist())
        child_ids = set(self.components.index.get_level_values(ID_CHILD).tolist())
        
        orphans = seqs - (parent_ids | child_ids)
        
        return orphans
        
    def get_orphans_components(self, level=ID_PARENT):
        """
        Get orphan ID 

        Returns
        -------
        List of ID        
        """
        ids = set(self.components.index.get_level_values(level).unique().tolist())
        seqs = set(self.sequences.index.unique().tolist())
        filters = self.components.index.get_level_values(level).isin(list(ids - seqs))
        return self.components.index[filters].tolist()
        
    def get_roots(self):
        """
        Get ID_CHILD roots of components and orphan ID sequences

        Returns
        -------
        List of ID        
        """
        root = ~self.sequences.index.isin(self.components.index.get_level_values(ID_CHILD))
        ids = self.sequences.index[root].unique().tolist() 
        return ids

    def get_leafs(self) -> List[int]:
        """
        Get ID_CHILD leafs of components 

        Returns
        -------
        List of ID_CHILD        
        """
        leaf = ~self.components.index.get_level_values(ID_CHILD).isin(self.components.index.get_level_values(ID_PARENT))
        return self.components[leaf].index.get_level_values(ID_CHILD).unique().tolist()

    def get_sequences(self, ids: int | List[int]) -> pd.DataFrame:
        """
        Get sequences of `ids`

        Returns
        -------
        A pandas DataFrame.        
        """
        return self.sequences.loc[self._id_list(ids), :]
    
    def get_components(self, pairs : Tuple[int, int] | List[Tuple[int, int]] = None) -> pd.DataFrame:
        """
        Get the  joint view of components and sequences of `pairs` (ID_PARENT, ID_CHILD)
        
        Returns
        -------
        A pandas DataFrame.        
        """
        comps = self.components.loc[pairs, :] if pairs is not None else self.components
        comps = comps.join(self.sequences, on=ID_CHILD, how='left')   
        comps = comps.join(self.sequences[KEYCODE], on=ID_PARENT, how='left', rsuffix='Parent')
        return comps

    def package_keys(self):
        return list(self._packages.keys())
    
    def set_package(self, key: str, value: List[Tuple[int, int]]):
        self._packages[key] = value
        self.param.trigger('notify_packages')

    def get_package(self, key: str) -> List[Tuple[int, int]]:
        if key not in self._packages:
            raise KeyError(f'DataSet.get_selections: {key} not in selections')
        return self._packages[key]

    def delete_all_packages(self):
        self._packages = {}
        self.param.trigger('notify_packages')

    def delete_package(self, key: str):
        if key not in self._packages:
            raise KeyError(f'DataSet.get_selections: {key} not in selections')
        del self._packages[key]
        self.param.trigger('notify_packages')
    
    
    def get_package_components(self, key: str, slope_resolution=None, param_cambium_estimation=None) -> pd.DataFrame:
        """
        Return the selection (a joint view of components and sequences) stored in dictonary `self.selection`.        
        
        Arguments
        ---------
        key: name of the selection.
        
        Returns
        -------
        A pandas DataFrame.
        """        
        def do_cambium_estimation(df, param):
            if CAMBIUM_LOWER not in df.columns:
                df[[CAMBIUM_LOWER, CAMBIUM_ESTIMATED, CAMBIUM_UPPER]] = [pd.NA, pd.NA, pd.NA]
            if (param.cambium_estimation_method != 'user values'):
                for id, row in df.iterrows():
                    df.loc[id, [CAMBIUM_LOWER, CAMBIUM_ESTIMATED, CAMBIUM_UPPER]] = cambium_estimation(param, row[CAMBIUM], row[BARK], row[SAPWOOD], row[DATA_VALUES])

        df = self.get_components(self.get_package(key))
        if slope_resolution is not None:
            df[SLOPE] = df[DATA_VALUES].apply(lambda x: slope(x, slope_resolution))
            
        if param_cambium_estimation is not None:
            do_cambium_estimation(df, param_cambium_estimation)
        
        return df
 
    def get_data(self, ids: Union[int, List[int], None] = None, category: Union[str, None] = None, 
                id_roots=None, max_depth: Union[int, None] = None) -> pd.DataFrame:
        """
        Create a joint view of components and sequences of `ids` descendants.        
        
        Arguments
        ---------
        ids: an `id` or a list of `id`.
        max_depth: if not `pd.NA`, limits the recursive loop to max_deep level.
        
        Returns
        -------
        A pandas DataFrame.
        """
        d = []
        if ids is None:
            ids = self.get_roots()
        ids = self._id_list(ids)
        if id_roots is None:
            include_parent = False
            id_roots = [-1] * len(ids)
        else:
            id_roots = self._id_list(id_roots)
            include_parent = True
        dict_id_roots = {id: id_root for id, id_root in zip(ids, id_roots)}
        
        pairs = set()
        tree = self.get_descendants(ids, max_depth=max_depth)        
        for node, offset in tree.filter(categories=category, max_depth=max_depth).items():
            id_parent = node.parent.id if node.id not in ids else dict_id_roots[node.id]
            if (node.id not in ids) or include_parent:
                if (node.id, id_parent) not in pairs:
                    d.append({ID_CHILD: node.id, ID_PARENT: id_parent, OFFSET: offset})
                    pairs.add((node.id, id_parent))
            
        components = pd.DataFrame(d, columns=components_index+components_cols)
        components.set_index(components_index, inplace=True, verify_integrity=True)  
        return components.join(self.sequences, on=ID_CHILD, how='left')

    def get_path_to_root(self, id: int) -> List[int]:
        if id is None:
            return []
        lst = [id]
        if id < 0:
            return lst
        id_parents = self.components.xs(id, level=ID_CHILD)
        if len(id_parents) > 0:
            if len(id_parents) > 1:
                logger.warning(f'multiple parents for {id} in {id_parents.index[0]}')
            lst = self.get_path_to_root(id_parents.index[0]) + lst
        #print('get_path_to_root:', lst)
        return lst
        
    def get_ascendants(self, id: int, recursive=False, categories=[MEAN, SET]):
        #perror('get_ascendants', id)
        comp = self.components
        id_parents = comp.xs(id, level=ID_CHILD).index.to_list() if id in comp.index.get_level_values(ID_CHILD) else []
        #perror('get_ascendants', id_parents, len(id_parents))
        l = self.sequences.loc[id_parents, CATEGORY].isin(categories).index.tolist() if len(id_parents) > 0 else []

        if recursive:
            for i in l:
                l += self.get_ascendants(i, recursive, categories)
        return l

    def get_category(self, category, mean_as_measure, check_mean_as_measure):
        #logger.info(f'get_category: {category} {mean_as_measure} {check_mean_as_measure}')
        if check_mean_as_measure and mean_as_measure and (category == MEAN):
            logger.warning(f'get_category: {category} is replaced by {MEASURE}')
            return MEASURE
        return category    

    def get_descendants(self, ids: int | List[int], max_depth=None, check_mean_as_measure=False) -> ComponentsTree():
        """
        Get descendants of `ids`.
        
        Arguments
        ---------
        ids: an `id` or a list of `id`.
        max_depth: if not `pd.NA`, limits the recursive loop to max_deep level.

        Returns
        -------
            A tree.
        """
        categories_keycodes = self.sequences.loc[:, [CATEGORY, MEANAS_MEASURE, KEYCODE]]
        data = self.components.join(categories_keycodes, on=ID_CHILD, how='left')
        group_parents = data.groupby(ID_PARENT)
        id_depth = []

        def iterate(parent, id, keycode, category, mean_as_measure, offset, depth, max_depth):
            if id in id_depth:
                keycode_id = self.sequences.at[id, KEYCODE]
                keycode_id_depth = [self.sequences.at[i, KEYCODE] for i in id_depth]                
                logger.warning(f'DataSet.get_descendants: circular reference: {id} in {id_depth}. Recursive loop aborded.\n {keycode_id} in {keycode_id_depth}')
                return None
            id_depth.append(id)
            category = self.get_category(category, mean_as_measure, check_mean_as_measure)
            node = ComponentsNode(parent, id, keycode, category, offset, depth=depth)
            if (id in group_parents.groups) and (category != MEASURE) and ((max_depth is None) or (depth+1 <= max_depth)):
                for (_, id_child), row in group_parents.get_group(id).iterrows():
                    child = iterate(node, id_child, row[KEYCODE], row[CATEGORY], row[MEANAS_MEASURE], row[OFFSET], depth+1, max_depth)
                    if child is not None:
                        node.append(child)
            id_depth.pop()
            return node
            #return node, errors
        
        tree = ComponentsTree()
        for id in self._id_list(ids):
            child = iterate(tree, id, categories_keycodes.at[id, KEYCODE], categories_keycodes.at[id, CATEGORY], categories_keycodes.at[id, MEANAS_MEASURE], 0, 0, max_depth)
            if child is not None:
                tree.append(child)
        return tree

    def edit_component(self, id_parent, id_child, value, notify=True, source=None):
        ids = [(id_parent, id_child)]
        self.freeze_components(ids)
        self.components.at[(id_parent, id_child), OFFSET] = np.round(value)

        self.log_components(ids, 'edit_component')
        if notify:
            self.notify_changes('notify_synchronize', source=source)
        if self.save_auto:
            self.dump()
    
    def edit_sequence(self, ids, column, value, notify=True, source=None):
        def check_keycode():
            if column == KEYCODE:
                if len(ids) > 1:
                    raise ValueError(f'Edit: multiple editions for {column} are not allowed')
                elif value in self.sequences[KEYCODE].values:
                    raise ValueError(f'Edit: duplicate {KEYCODE} for value: {value}')
        
        def update(ids, column, value):
            #perror('update', ids, column, value)
            for id in ids:
                self.sequences.at[id, column] = value       
            
        def update_date_end(ids):
            self.sequences.loc[ids, DATE_END] = self.sequences.loc[ids, DATE_BEGIN] + self.sequences.loc[ids, DATA_LENGTH] - 1

        def update_date_begin(ids):
            self.sequences.loc[ids, DATE_BEGIN] = self.sequences.loc[ids, DATE_END] - self.sequences.loc[ids, DATA_LENGTH] + 1
            
        if sequences_dtype_dict[column].lower().startswith('int'):
            value = np.round(value) if pd.notna(value) else pd.NA
        elif sequences_dtype_dict[column].lower().startswith('float'):
            if pd.isna(value):
                value = pd.NA   
        elif sequences_dtype_dict[column].lower().startswith('bool'):
            value = string_to_boolean(value)
        
        ids = self._id_list(ids)    
        check_keycode() 
        self.freeze_sequences(ids)
        
        update(ids, column, value)
        if column == DATA_VALUES:
            update(ids, DATA_LENGTH, len(value))
            update_date_end(ids)
        elif column == DATE_BEGIN:
            update_date_end(ids)
        elif column == DATE_END:
            update_date_begin(ids)

        self.log_sequences(ids, 'edit_sequence')
        
        if notify:
            self.notify_changes('notify_synchronize', source=source)
        if self.save_auto:
            self.dump()

    def set_inconsistent_on_ascendants(self, id, notify=True, source=None):
        ascendants = self.get_ascendants(id, recursive=True)
        if len(ascendants) > 0:
            self.edit_sequence(ascendants, INCONSISTENT, True, notify=False)
        if notify:
            self.notify_changes('notify_synchronize', source=source)

    def shift_offsets(self, parent_id, child_ids=None, notify=True):
        """
        Get children of `parent_id` and shift the children offsets to 0.
        """
        data = self.get_data(parent_id, max_depth=1)
        if child_ids is not None:
            data = data[data.index.get_level_values(ID_CHILD).isin(child_ids)]
        if data[OFFSET].isna().any():
            raise ValueError(f'DataSet.shift_offsets: one or more ({ID_PARENT}, {ID_CHILD}) contain NA values in {OFFSET} field.')
        ids = data.index.to_list()
        
        self.freeze_components(ids)
        self.components.loc[ids, OFFSET] -= data[OFFSET].min()
        self.log_components(ids, 'shift_offsets')
        if notify:
            self.notify_changes('notify_synchronize')

    def copy_dates_to_offsets(self, parent_id, child_ids=None, notify=True):
        """
        Get children of `parent_id`, copy dates to offsets and shift to 0 (if `shift` is True).
        """
        data = self.get_data(parent_id, max_depth=1)
        if child_ids is not None:
            data = data[data.index.get_level_values(ID_CHILD).isin(child_ids)]
        if data[DATE_BEGIN].isna().any():
            logger.warning(f'one or more ({ID_PARENT}, {ID_CHILD}) contain NA values in {DATE_BEGIN} field.')
        ids = data.index.to_list()
        self.freeze_components(ids)
        self.components.loc[ids, OFFSET] = data.loc[ids, DATE_BEGIN]
        self.log_components(ids, 'copy_dates_to_offsets')
        if notify:
            self.notify_changes('notify_synchronize')

    def set_offsets_to_dates(self, parent_id, child_ids=None, notify=True):
        """
        Get children of `parent_id`, copy dates to offsets and shift to 0 (if `shift` is True).
        """
        data = self.get_data(parent_id, max_depth=1)
        if child_ids is not None:
            data = data[data.index.get_level_values(ID_CHILD).isin(child_ids)]
        if data[OFFSET].isna().any():
            raise ValueError(f'DataSet.set_offsets_to_dates: one or more ({ID_PARENT}, {ID_CHILD}) contain NA values in {OFFSET} field.')
        min_offset = data[OFFSET].min()
        min_date = data.at[(data[OFFSET] == min_offset).idxmax(), DATE_BEGIN]
        if pd.isna(min_date):
            raise ValueError(f'DataSet.set_offsets_to_dates: {DATE_BEGIN} corresponding to min {OFFSET} contains NA value.')
        data[OFFSET] -= min_offset
        ids = data.index.get_level_values(ID_CHILD).to_list()
        self.freeze_sequences(ids)
        self.sequences.loc[ids, DATE_BEGIN] =  data.reset_index().set_index(ID_CHILD)[OFFSET] + min_date
        self.sequences.loc[ids, DATE_END] = self.sequences.loc[ids, DATE_BEGIN] + self.sequences.loc[ids, DATA_LENGTH] - 1        
        self.log_sequences(ids, 'set_offsets_to_dates')
        if notify:
            self.notify_changes('notify_synchronize')

    def set_date_begin(self, parent_id, child_ids=None, notify=True):
        data = self.get_data(parent_id, max_depth=1)
        if child_ids is not None:
            data = data[data.index.get_level_values(ID_CHILD).isin(child_ids)]
        if data[DATE_END].isna().any():
            raise ValueError(f'DataSet.set_date_begin: one or more ({ID_PARENT}, {ID_CHILD}) contain NA values in {DATE_END} field.')
        if data[DATA_LENGTH].isna().any():
            raise ValueError(f'DataSet.set_date_begin: one or more ({ID_PARENT}, {ID_CHILD}) contain NA values in {DATA_LENGTH} field.')
        ids = data.index.get_level_values(ID_CHILD).to_list()
        self.freeze_sequences(ids)
        self.sequences.loc[ids, DATE_BEGIN] =  self.sequences.loc[ids, DATE_END] - self.sequences.loc[ids, DATA_LENGTH] + 1
        self.log_sequences(ids, 'set_date_begin')
        if notify:
            self.notify_changes('notify_synchronize')

    def set_date_end(self, parent_id, child_ids=None, notify=True):
        data = self.get_data(parent_id, max_depth=1)
        if child_ids is not None:
            data = data[data.index.get_level_values(ID_CHILD).isin(child_ids)]
        if data[DATE_BEGIN].isna().any():
            raise ValueError(f'DataSet.set_date_end: one or more ({ID_PARENT}, {ID_CHILD}) contain NA values in {DATE_BEGIN} field.')
        if data[DATA_LENGTH].isna().any():
            raise ValueError(f'DataSet.set_date_end: one or more ({ID_PARENT}, {ID_CHILD}) contain NA values in {DATA_LENGTH} field.')
        ids = data.index.get_level_values(ID_CHILD).to_list()
        self.freeze_sequences(ids)
        self.sequences.loc[ids, DATE_END] =  self.sequences.loc[ids, DATE_BEGIN] + self.sequences.loc[ids, DATA_LENGTH] - 1
        self.log_sequences(ids, 'set_date_end')
        if notify:
            self.notify_changes('notify_synchronize')


    def check_ring_count(self, parent_id):
        """
        Check the ring count of the given parent index.

        Args:
            parent_id (int): The index of the parent.

        Raises:
            ValueError: If the ring count does not match the length of the values.

        Returns:
            None
        """
        def length_(values):
            l = len(values) if values is not None else 0
            #logger.debug(f'{l}, {values}')
            return l
        data = self.get_data(parent_id, max_depth=1)
        ring_count = data[DATA_VALUES].apply(lambda x: length_(x))
        if not (ring_count == data[DATA_LENGTH]).all():
            raise ValueError(f'{DATA_LENGTH} does not match the length of {DATA_VALUES}.')

    def check_date_ring_count(self, id):
        """
        Checks the date and ring count consistency for a given parent index.

        Args:
            parent_id (int): The index of the parent node.

        Raises:
            ValueError: If the date and ring count are not consistent.

        Returns:
            None
        """
        self.check_ring_count(id)
        data = self.get_data(id, max_depth=1)
        min_date = data[DATE_BEGIN].min()
        data[DATE_BEGIN] -= min_date        
        data[DATE_END] -= min_date
        if not ((data[DATE_END] - data[DATE_BEGIN] + 1) == data[DATA_LENGTH]).all():
            for row in data.itertuples():
                db = row.DATE_BEGIN
                de = row.DATE_END
                dl = row.DATA_LENGTH
                dl2 = de - db + 1
                if dl2 != dl:
                    logger.warning(f' {row.KEYCODE} {row.DATE_BEGIN} {row.DATE_END} {row.DATA_LENGTH} {row.OFFSET} {dl2} != {dl}')
            raise ValueError(f'DataSet.check_date_ring_count: {DATE_BEGIN} / {DATE_END} and {DATA_LENGTH} are not consistent.')
    
    def check_offset_begin_date(self, id):
        """
        Check the consistency between the 'OFFSET' and 'DATE_BEGIN' columns in the dataset.

        Args:
            parent_id (int): The index of the parent dataset.

        Raises:
            ValueError: If the 'OFFSET' and 'DATE_BEGIN' columns are not consistent.

        Returns:
            None
        """
        data = self.get_data(id, max_depth=1)
        data[OFFSET] -= data[OFFSET].min()
        min_date = data[DATE_BEGIN].min()
        data[DATE_BEGIN] -= min_date
        if not(data[OFFSET] == data[DATE_BEGIN]).all():
            raise ValueError(f'DataSet.check_offset_begin_date: {DATE_BEGIN} and {OFFSET} are not consistent.')
        
    def check_date_offset_count(self, id):
        """
        Checks the date offset count for a given parent index.

        Parameters:
        - parent_id (int): The index of the parent.

        Returns:
        None
        """
        self.check_date_ring_count(id)
        self.check_offset_begin_date(id)

    def set_dates(self, id, date_begin, data_length=None, sequences=None, warning=True, notify=True, source=None):
        """
        Set DATE_END and DATE_BEGIN (if not NA) of `id` series given a `date_begin` 
        and a `ring_count`.

        Parameters:
        - id (int): The index of the series.
        - date_begin (datetime): The beginning date for the series.
        - data_length (int, optional): The length of the data. If not provided, it will be retrieved from the sequences.
        - sequences (pandas.DataFrame, optional): The sequences dataframe. If not provided, it will use the self.sequences.
        - warning (bool, optional): Whether to show a warning if there is a potential inconsistent mean. Default is True.
        """
        if sequences is None:
            sequences = self.sequences
        if pd.notna(date_begin):
            keycode = sequences.at[id, KEYCODE]
            date_first = sequences.at[id, DATE_BEGIN]
            if pd.notna(date_first) and (date_begin != date_first) and (warning):
                logger.warning(f'potential inconsistent mean, {keycode} {DATE_BEGIN} changed: {date_begin} ')
            self.freeze_sequences(id)
            if data_length is None:
                data_length = sequences.at[id, DATA_LENGTH]
            else:
                sequences.at[id, DATA_LENGTH] = data_length
            sequences.at[id, DATE_BEGIN] = date_begin                
            sequences.at[id, DATE_END] = date_begin + data_length - 1
            self.log_sequences(id, 'set_dates')
            if notify:
                self.notify_changes('notify_synchronize', source=source)

            
    def set_chononology_info(self, id, means, weights, offsets, signatures, data_type, data_samples, sequences=None):
        """
        Set the mean information for a given index in the dataset.

        Args:
            id (int): The index of the sequence to update.
            means (list): The mean values of the sequence.
            weights (list): The weights of the sequence.
            offsets (list): The offsets of the sequence.
            data_type (str): The type of data.
            data_samples (pandas.DataFrame): The data samples.
            sequences (pandas.DataFrame, optional): The sequences dataframe. Defaults to None.

        Returns:
            None
        """
        if sequences is None:
            sequences = self.sequences
        #self.freeze_sequences(id)
        sequences.at[id, DATA_VALUES] = means
        sequences.at[id, DATA_TYPE] = data_type
        sequences.at[id, DATA_LENGTH] = len(means)
        sequences.at[id, DATA_WEIGHTS] = weights
        sequences.at[id, DATA_INFO] = offsets
        sequences.at[id, DATA_SIGNATURES] = signatures
        sequences.at[id, DATE_SAMPLING] = datetime.now()
        sequences.at[id, INCONSISTENT] = False
        sequences.at[id, CAMBIUM] = False
        sequences.at[id, PITH] = False
        sequences.at[id, CAMBIUM_SEASON] = ''
        sequences.at[id, SAPWOOD] = pd.NA
        
        sequences.at[id, CATEGORY] = MEAN
        for key in [SITE_ELEVATION, SITE_CODE, SITE_LATITUDE, SITE_LONGITUDE, SPECIES, LABORATORY_CODE, PROJECT, URI]: 
            l = data_samples[key].unique()
            if len(l) == 1:
                sequences.at[id, key] = l[0]
        date_min = data_samples[DATE_BEGIN].min()
        self.set_dates(id, date_min, len(means), notify=False)
        sequences.at[id, COMPONENT_COUNT] = len(data_samples)
        
        #self.log_sequences(id, 'Mean update')
    
    def means(self, ids, date_as_offset=False, biweight=False, num_threads=1, check_mean_as_measure=False, slope_resolution=0, min_elements=4):
        """
        Compute means for the given indices.

        Args:
            ids (list): List of indices for which to compute means.
            date_as_offset (bool, optional): Whether to treat dates as offsets. Defaults to False.
            biweight (bool, optional): Whether to use biweight location and scale estimators. Defaults to False.
            num_threads (int, optional): Number of threads to use for computation. Defaults to 1.

        Returns:
            None
        """
        tree = self.get_descendants(ids, check_mean_as_measure=check_mean_as_measure)
        
        node_means = tree.filter(categories=[MEAN, SET], max_depth=1) 
        id_means = []
        data_dict = {}
        for node in node_means:
            id = node.id
            if id in ids: # Need in the return
                if (id not in id_means):  # Never computed 
                    id_means.append(id)
                    samples = {node.id: offset for node, offset in node.descendants[MEASURE].items()}
                    dt_data = self.get_sequences(list(samples.keys())).copy()
                    dt_data[OFFSET] = list(samples.values())
                    data_dict[id] = dt_data
        
        results = compute_means(data_dict, ring_type=RAW, date_as_offset=date_as_offset, biweight=biweight, num_threads=num_threads, slope_resolution=slope_resolution, min_elements=min_elements)
        for id, values in results.items():
            means, weights, offsets, signatures = values
            self.set_chononology_info(id, means, weights, offsets, signatures, RAW, data_dict[id], sequences=None)
            self.set_inconsistent_on_ascendants(id, notify=False)           

        if self.save_auto:
            self.dump()

    def detrend(self, ids, ring_type, window_size=5, do_log=False, date_as_offset=False, biweight=False, slope_resolution=0, min_elements=4):
        """
        Detrends the data for the specified indices.

        Args:
            ids (list): List of indices to detrend.
            ring_type (str): Type of ring data.
            window_size (int, optional): Size of the moving window for detrending. Defaults to 5.
            do_log (bool, optional): Whether to apply logarithmic transformation. Defaults to False.
            date_as_offset (bool, optional): Whether to treat dates as offsets. Defaults to False.
            biweight (bool, optional): Whether to use biweight location and scale estimators. Defaults to False.
            num_threads (int, optional): Number of threads to use for parallel processing. Defaults to 1.

        Returns:
            pandas.DataFrame: Detrended data for the specified indices.
        """
        tree = self.get_descendants(ids)
        num_threads=1
        
        if MEASURE in tree.descendants:
            ids_samples = list(set([node.id for node in tree.descendants[MEASURE].keys()]))
            data_samples = self.get_sequences(ids_samples)
            dt_samples = detrend(data_samples, ring_type, window_size=window_size, do_log=do_log, num_threads=num_threads)
        dt_chonology = []
        if MEAN in tree.descendants:
            node_means = tree.filter(categories=[MEAN], max_depth=1)
            id_means = []
            data_dict = {}
            for node in node_means:
                id = node.id
                if id in ids: # Need in the return
                    if (id not in id_means):  # Never computed 
                        id_means.append(id)
                        samples = {node.id: offset for node, offset in node.descendants[MEASURE].items()}
                        dt_data = dt_samples.loc[list(samples.keys()), :]
                        dt_data[OFFSET] = list(samples.values())
                        data_dict[id] = dt_data
            results = compute_means(data_dict, ring_type=ring_type, date_as_offset=date_as_offset, biweight=biweight, num_threads=num_threads, slope_resolution=slope_resolution, min_elements=4)
            for id, values in results.items():
                means, weights, offsets, signatures = values
                seq = self.get_sequences(id).copy()
                self.set_chononology_info(id, means, weights, offsets, signatures, ring_type, data_dict[id], sequences=seq)
                dt_chonology.append(seq)
        with warnings.catch_warnings():
            # TODO: pandas 2.1.0 has a FutureWarning for concatenating DataFrames with Null entries
            warnings.filterwarnings("ignore", category=FutureWarning)
            data_dt = pd.concat([dt_samples]+dt_chonology)
        return data_dt.loc[ids, :]

    def check(self, id):
        def check_children(node):
            children = node.get_children()
            id_children = [node.id for node in children.keys()]
            category_children = [node.category for node in children.keys()]
            offsets = pd.Series([offset for offset in children.values()], index=id_children)
            offset_nonan = offsets.notna()
            noffset = offset_nonan.sum()
            offsets_norm = offsets - offsets.min()
            keycodes = self.sequences.loc[id_children, KEYCODE]
            dates = self.sequences.loc[id_children, DATE_BEGIN]
            date_nonan = dates.notna()
            dates_norm = dates - dates.min()
            ndate = date_nonan.sum()
            norm = dates_norm.max()+offsets_norm.max()
            diff =  dates_norm.fillna(norm) != offsets_norm.fillna(norm) 
            equal =  dates_norm.fillna(norm) == offsets_norm.fillna(norm) 
            ndiff = diff.sum()
            if node.category == MEAN:
                if (ndiff == 0) and (ndate == len(dates)) and (noffset == len(offsets)): 
                    if SET not in category_children:
                        msg = '1: dates and offsets are consistent.'
                    else:
                        msg = '-1: dates and offsets are consistent. But some children are "set".'
                elif (ndate == len(dates)) and (noffset == 0):
                    # all dates, no offset
                    if SET not in category_children:
                        msg = '2: dates are available, no offsets. Offsets update required.'
                    else:
                        msg = '-2: dates are available, no offsets. Offsets update required. But some children are "set".'
                elif (noffset == len(offsets)) and (ndate == 0):
                    # all offsets, no date
                    msg = '3: offsets are available, no dates. Undated serie.'
                elif (noffset == len(offsets)) and (equal[dates != pd.NA].sum() == ndate):
                    # all offsets, subset of dates and offsets consistent
                    if SET not in category_children:
                        msg = '4: offsets are available, some empty dates. subset of dates and offsets are consistent. Years update required.'
                    else:
                        msg = '-4: offsets are available, some empty dates. But some children are "set".'
                elif (ndate == len(dates)) and (equal[offsets != pd.NA].sum() == noffset):
                    # all offsets, subset of dates and offsets consistent
                    if SET not in category_children:
                        msg = '5: Years are available, some empty offsets. subset of dates and offsets are consistent. Offsets update required.'
                    else:
                        msg = '-5: Years are available, some empty offset. But some children are "set".'
                else:
                    if SET not in category_children:
                        msg = '-6: Years and offsets are unconsistent. Undated serie. '
                    else:
                        msg = f'-7: Contain {SET}, dates and offsets are unconsistent. Undated serie.'
            else:
                msg = f'-7: Years and offsets are unconsistent. Parent is not a {MEAN}. Correction required'
            
            info = [(node.id, self.sequences.at[node.id, KEYCODE], self.sequences.at[node.id, CATEGORY],
                          self.sequences.at[node.id, DATE_BEGIN],  pd.NA,  pd.NA , pd.NA, pd.NA)]
            info += zip(id_children, keycodes.to_list(), category_children, dates.tolist(), offsets.tolist(), 
                        dates_norm.tolist(), offsets_norm.tolist(), equal)            
            df = pd.DataFrame(info, columns=[ID, KEYCODE, CATEGORY, DATE_BEGIN, OFFSET, DATE_BEGIN_NORM, OFFSET_NORM, 'date ~ offset'])          
            return (msg, df)
        
        def get(node):
            out[node.id] = check_children(node)
            for child_node in node.children:
                if child_node.category != MEASURE:
                    get(child_node)
        
        out = {}
        tree = self.get_descendants(id)
        for child_node in tree.children:
            get(child_node)        
        return out

    def get_keycodes(self, ids, fields=None):
            """
            Generate unique keycodes for each element stored in attributes of the Sequences DataFrame.

            Args:
                ids (list): A list of indices corresponding to the elements for which keycodes need to be generated.
                fields (list, optional): A list of fields to include in the generated keycodes. Defaults to None.

            Returns:
                dict: A dictionary containing the keycodes as keys and their corresponding values.
            """
            cols = [KEYCODE, PROJECT]
            if fields is not None:
                cols += fields
            data = self.sequences.loc[ids, cols]
            if fields == None:
                if len(data) == len(data[KEYCODE].unique()):
                    return {x:y for x, y in zip(data.index, data[KEYCODE])}
                if len(data) == len(data[[PROJECT, KEYCODE]].drop_duplicates()):
                    return {x:f'{x}/{y}/{z}' for x, y, z in zip(data.index, data[PROJECT], data[KEYCODE])}
                return {x:f'{x}/{y}' for x, y in zip(data.index, data[KEYCODE])}
            else:
                return {x:f'{x}/{y}' for x, y in zip(data.index, data[fields])}

    def detrend_package(self, package_key, ring_type, window_size=5, do_log=False, date_as_offset=False, biweight=False, 
                       slope_resolution=None, cambium_estimation=None):
        
        def info(df, ids):
            message_type = 'primary'
            message = f'Detrend data is {ring_type} '
            message += ', '.join([f'{index}: {valeur}' for index, valeur in df[CATEGORY].value_counts().items()]) +'.'
            if len(ids) != len(df[ID_CHILD]):
                message += 'Duplicate series.'
            if df[INCONSISTENT].any():
                message += ' one or more series is inconsistent.'
                message_type = 'warning'

            return message_type, message

        num_threads=1
        data = self.get_package_components(package_key, slope_resolution, cambium_estimation).reset_index()
        ids = data[ID_CHILD].unique().tolist()
        if len(ids) <= 0:
            return None, 'warning', 'Detrend data is empty set'
        elif (ring_type != RAW):
            dt_data = self.detrend(ids, ring_type, window_size=window_size, do_log=do_log, date_as_offset=date_as_offset, biweight=biweight, num_threads=num_threads)
            data_cols = [DATA_LENGTH, DATA_TYPE, DATA_VALUES, DATA_WEIGHTS, DATA_INFO, DATA_SIGNATURES, INCONSISTENT]
            other_cols = data.columns.difference(data_cols)
            data = data[other_cols].join(dt_data[data_cols], on=ID_CHILD, how='left')

        message_type, message = info(data, ids)
        return data, message_type, message

    def read_buffer(self, keycode_parent, buffer, mine_type=None):
        #perror('read_buffer', mine_type)
        suffix = '.p'
        if mine_type == 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet':
            suffix = '.xlsx'
        elif mine_type == 'text/csv':
            suffix = '.csv'
        elif mine_type == 'application/json':
            suffix = '.json'
        output_file = Path(self.cfg_tmp) / Path(keycode_parent).with_suffix(suffix)
        with open(output_file, 'wb') as f:
            f.write(buffer)  # Récupère le contenu binaire du buffer
        
        
        self.load(output_file) 
        output_file.unlink()
        return self   

def string_to_boolean(value):
    if isinstance(value, str):
        if value.lower() == 'true':
            return True
        elif value.lower() == 'false':
            return False
        else:
            return pd.NA
    else:
        return value
 



"""
    Ploter module
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
 
import numpy as np
import pandas as pd

import param
import panel as pn
from panel.viewable import Viewer

from bokeh.plotting import figure
from bokeh.models import ColumnDataSource, FixedTicker, Label, Range1d, BoxAnnotation
from bokeh.palettes import Category20, Category10
from bokeh.core.enums import LegendLocation

from pyDendron.dataname import *
from pyDendron.app_logger import logger, perror

class Ploter(Viewer):
    ANATOMY = 'Anatomy'
    ANATOMY_COLOR = {HEARTWOOD: 'black', 
                     PITH: 'blue', 
                     SAPWOOD:'red', 
                     CAMBIUM:'red',
                     BARK: 'red',
                     'signature_90': 'blue',
                     'signature_75': 'green'}


    width = param.Integer(default=1000, bounds=(50, 4000), step=10)
    height = param.Integer(default=1000, bounds=(50, 4000), step=10)
    figure_title = param.String(' ')
    figure_title_font_size = param.Integer(default=14, bounds=(1, 40), step=1)

    data_type = param.Selector(default='Raw', label='Data type *', objects=['Raw', 'Log', 'Detrend'], doc='where is the data')
    color = param.Selector(default=ANATOMY, objects=['None', KEYCODE, ANATOMY], 
            doc=f'None: all black, {KEYCODE}: one color per {KEYCODE}, {ANATOMY}: color pith, sapwood... ')
    draw_type = param.Selector(default='Line', label='Draw type *', objects=['Line', 'Step', 'Timeline'], doc='Line draw type.') 
    group_by = param.Selector(default=None, label='Group by *', objects=[None, TAG, CATEGORY, KEYCODE, DATE_END], doc='group by')
    #draw_selection = param.Boolean(False, label='Draw selection', doc='Draw only selected series in package')
    
    x_offset_mode = param.Selector(default=DATE_BEGIN, label='X offset mode *', objects=['None', DATE_BEGIN, OFFSET], doc='Delay of the rings')
    y_offset_mode = param.Selector(default='Stack', label='Y offset mode *', objects=['Zero', 'Stack'], doc='y curve position')
    y_height = param.Selector(default='[0, max]', label='Y Height *', objects=['[0, max]', '[min, max]'], doc='height of a serie')
    y_delta = param.Number(default=0, label='Y Delta *', doc='marge between two series')
    x_range_step = param.Integer(default=25, bounds=(5, 200), step=5)
    axis_marge = param.Integer(default=35, bounds=(0, 500), step=5)
    axis_font_size = param.Integer(default=10, bounds=(1, 20), step=1)

    cambium_estimation = param.Boolean(True, label= 'Cambium estimation *', doc='Draw cambium estimation')
    show_dates = param.Boolean(True, doc='Add dates at the start and end of the serie')
    show_estimated_dates = param.Boolean(True, doc='Add estimated dates at the end of the serie')
    show_cambium_season = param.Boolean(True, doc='Add cambium season at the end of the serie')

    curve_axe = param.Selector(default='1 mm', label='Curve axe *', objects=['None', '1 mm', '0 mm', 'Mean'], doc=f'position of {KEYCODE} axe')
    anatomy = param.Selector(default='Axe', label='Anatomy *', objects=['Curve', 'Axe'], doc='color of the curve')
    
    line_width_tree = param.Number(default=0.5, bounds=(0.25, 4.0), step=0.25)
    line_width_mean = param.Number(default=1, bounds=(0.25, 4.0), step=0.25)
    marker_size = param.Integer(default=1, bounds=(1, 100), step=1)

    legend = param.Selector(default=KEYCODE, label='Legend *', objects=[KEYCODE, 'Number', 'Number+' + KEYCODE, 'None'], doc=f'legend type')
    legend_location = param.Selector(default='Curve begin', label='Legend location *', objects=['None', 'Axe Y', 'Curve begin', 'Curve end']+[loc for loc in LegendLocation], doc='Legend location')
    legend_font_size = param.Integer(default=10, bounds=(1, 20), step=1)
    #grid_line_visible = param.Boolean(False, doc='Show grid line')

    signature_75 = param.Boolean(False, doc='Signature 75')
    signature_90 = param.Boolean(False, doc='Signature 90')
    
    border = param.Boolean(False, doc='Log transform')
    
    def __init__(self, ploter_name='ploter', **params):
        super(Ploter, self).__init__(**params)   
        self.ploter_name = ploter_name
        self.draw_data = None
        self.draw_group = None 
        self.data = None
        self.x_range, self.y_range = None, None
        self.figure_pane = pn.pane.Bokeh(height=self.height, width=self.width)
        self.clear()
        self._layout = self.figure_pane
        perror('Ploter init', self.width, self.height)
 
    def __panel__(self):
        return self._layout
                
    #@param.depends("width", watch=True)
    def _update_width(self):
        self.figure_pane.width = self.width

    #@param.depends("height", watch=True)
    def _update_height(self):
        self.figure_pane.height = self.height
    
    def get_pith_optimun(self, data_len):
        return int(data_len*0.1)
 
    def prepare_data(self, data):
        self.data = data

        def init_ColumnDataSource():
            return ColumnDataSource()

        def get_x_offset(row):
            if self.x_offset_mode == 'None':
                return 0
            elif self.x_offset_mode == DATE_BEGIN:
                return row[DATE_BEGIN]
            return row[OFFSET]

        def get_y_offset(row, cum_y_offset):
            data = row[DATA_VALUES]
            v = self.y_delta
            if self.draw_type == 'Timeline':
                v += 50
            else:
                v += np.nanmax(data) if self.y_offset_mode == 'Stack' else 0
            cum_y_offset += v
            return v, cum_y_offset
        
        def get_values(row, info):
            values = row[DATA_VALUES]

            sapwood_offset = row[SAPWOOD]
            info[SAPWOOD] = init_ColumnDataSource()
            info[HEARTWOOD] = init_ColumnDataSource()

            if pd.isna(sapwood_offset) or sapwood_offset < 0:
                sapwood_offset = len(values) - 1
                info['sapwood_offset'] = sapwood_offset 
            info[HEARTWOOD].data['x'] = np.arange(0, sapwood_offset + 1) + info['x_offset']
            
            info[HEARTWOOD].data['w'] = values[:sapwood_offset + 1]
            info[HEARTWOOD].data['y'] = info[HEARTWOOD].data['w'] + info['y_offset'] 
            
            info['is_sapwood'] = not(pd.isna(sapwood_offset) or sapwood_offset < 0)
            info[SAPWOOD].data['x'] = np.arange(sapwood_offset, len(values)) + info['x_offset']
            info[SAPWOOD].data['w'] = values[sapwood_offset:]
            info[SAPWOOD].data['y'] = info[SAPWOOD].data['w'] + info['y_offset']
            
        def get_missing_ring(row, info):
            values = row[DATA_VALUES]
            begin = np.where(~np.isnan(values))[0][0]
            end = np.where(~np.isnan(values))[0][-1]
            
            if begin > 0:
                info[MISSING_RING_BEGIN] = init_ColumnDataSource()
                info[MISSING_RING_BEGIN].data['x'] = np.arange(0, begin) + info['x_offset']
                info[MISSING_RING_BEGIN].data['w'] = [values[begin]] * begin
                info[MISSING_RING_BEGIN].data['y'] = [values[begin] + info['y_offset']] * begin 
            if end < len(values):
                info[MISSING_RING_END] = init_ColumnDataSource()
                info[MISSING_RING_END].data['x'] = np.arange(end, len(values)) + info['x_offset']
                info[MISSING_RING_END].data['w'] = [values[end]] * (len(values) - end)
                info[MISSING_RING_END].data['y'] = [values[end] + info['y_offset']] * (len(values) - end) 
            
        def get_pith(row, info):
            values = row[DATA_VALUES] 
            x_min = info['x_offset']
            i = np.where(~np.isnan(values))[0][0]
            w = values[i]
            info[PITH] = init_ColumnDataSource()
            if pd.notna(row[PITH]) and row[PITH]:
                info[PITH].data['x'] = [info['x_offset']]
                info[PITH].data['w'] = [w]
                info[PITH].data['y'] = [get_y(info, w)]
            return x_min
                
        def get_cambium(row, info):
            info['is_cambium_estimated'] = False
            values = row[DATA_VALUES]
            x = len(values) - 1
            x_max = x + info['x_offset']

            w = values[np.where(~np.isnan(values))[0][-1]]
            info[CAMBIUM] = init_ColumnDataSource()
            info[CAMBIUM_ESTIMATED] = init_ColumnDataSource()
            info[CAMBIUM_BOUNDARIES] = init_ColumnDataSource()
            if pd.notna(row[CAMBIUM]) and row[CAMBIUM]:
                info[CAMBIUM].data['x'] = [x + info['x_offset']]
                info[CAMBIUM].data['w'] = ['NA']
                info[CAMBIUM].data['y'] = [get_y(info, w)]
            else:
                if (CAMBIUM_ESTIMATED in row) and row[[CAMBIUM_ESTIMATED, CAMBIUM_LOWER, CAMBIUM_UPPER]].notna().all():
                    info['is_cambium_estimated'] = True

                    lower, estimated, upper = row[CAMBIUM_LOWER], row[CAMBIUM_ESTIMATED], row[CAMBIUM_UPPER]
                    wo = get_y(info, w)
                    xe = estimated + info['x_offset']
                    xl = lower + info['x_offset']
                    xu = upper + info['x_offset']
                    
                    diff_h = (np.nanmax(values) - min(np.nanmin(values), 0)) / 5
                    h = 25 if diff_h > 25 else diff_h
                    h = 0.5 if h < 0.5 else h
                    #perror(f'---> {row[KEYCODE]} {diff_h} {h} {np.nanmax(values)} {np.nanmin(values)}')
                    #h = d if d < 25 else 25
                    info[CAMBIUM_BOUNDARIES].data['x'] = np.arange(xl, xu+1) 
                    info[CAMBIUM_BOUNDARIES].data['w'] = np.array(['NA']*(xu-xl+1))
                    info[CAMBIUM_BOUNDARIES].data['y'] =  np.array([get_y(info, w)]*(xu-xl+1))
                    
                    info[CAMBIUM_ESTIMATED].data['x0'] = [xu, xe, xl]
                    info[CAMBIUM_ESTIMATED].data['x1'] = [xu, xe, xl]
                    info[CAMBIUM_ESTIMATED].data['w'] = ['NA'] * 3
                    info[CAMBIUM_ESTIMATED].data['y0'] = [wo - h] * 3
                    info[CAMBIUM_ESTIMATED].data['y1'] = [wo + h] * 3
                    x_max = xu if xu > x_max else x_max
                elif (CAMBIUM_ESTIMATED in row) and row[[CAMBIUM_ESTIMATED, CAMBIUM_LOWER, CAMBIUM_UPPER]].notna().any():
                    logger.warning(f'Incomplet cambium estimation for {row[KEYCODE]}')
                    
            return x_max

        def get_bark(row, info):
            values = row[DATA_VALUES]
            x = len(values)
            w = values[np.where(~np.isnan(values))[-1]]
            info[BARK] = init_ColumnDataSource()
            if pd.notna(row[BARK]) and row[BARK]:
                info[BARK].data['x'] = [x + info['x_offset']]
                info[BARK].data['w'] = [w]
                info[BARK].data['y'] = [get_y(info, w)]
                
        def get_border(row, info):
            info['border'] = init_ColumnDataSource()
            info['border'].data['top'] = [info['y_max']]
            info['border'].data['bottom'] = [info['y_min']]
            info['border'].data['left'] = [info['x_min']]
            info['border'].data['right'] = [info['x_max']]
            
        def get_min_max(row, info):
            info['x_min'] = get_pith(row, info)
            info['x_max'] = get_cambium(row, info)
            info['w_min'] = np.nanmin(row[DATA_VALUES])
            info['w_max'] = np.nanmax(row[DATA_VALUES])
            info['w_mean'] = np.nanmean(row[DATA_VALUES])
            info['y_min'] = info['y_offset'] #info['w_min'] + info['y_offset']
            info['y_max'] = next_cum_y_offset
            info['y_mean'] = info['w_mean'] + info['y_offset']
            info['y_label'] = round(info['y_mean'], 3)    
        
        def get_text(row, i, info):
            def add_text(value, text):
                if (value is None) or pd.isna(value):
                    return text
                if text != '':
                    text += ' \u2014 '
                return text + f'{value}'
            
            info['text_begin'] = init_ColumnDataSource()
            info['text_begin'].data['x'] = [info['x_min']]
            info['text_begin'].data['y'] = [info['y_mean']]
            info['text_end'] = init_ColumnDataSource()
            info['text_end'].data['x'] = [info['x_max']]
            info['text_end'].data['y'] = [info['y_mean']]
            text_begin = ''
            text_end = ''
            
            keycode = self.get_legend(i, info[KEYCODE])
            if self.legend_location == 'Curve begin':
                text_begin = add_text(keycode, text_begin)
            elif self.legend_location == 'Curve end':
                text_end = add_text(keycode, text_end)
            if self.show_dates :
                text_begin = add_text(row[DATE_BEGIN], text_begin)
                text_end = add_text(row[DATE_END], text_end)
            if self.show_estimated_dates and info['is_cambium_estimated']:
                b = row[DATE_BEGIN]
                lower, estimated, upper = b + row[CAMBIUM_LOWER], b + row[CAMBIUM_ESTIMATED], b + row[CAMBIUM_UPPER]
                text_end += f' [{lower}, {estimated}, {upper}]'
            if self.show_cambium_season and (pd.notna(row[CAMBIUM_SEASON]) and (row[CAMBIUM_SEASON] != '')):
                text_end = add_text(f'\U0001FA93 {row[CAMBIUM_SEASON]}', text_end)
            
            info['text_begin'].data['text'] = [text_begin]
            info['text_end'].data['text'] = [text_end]
        
        def get_curve_axe(row, info):
            info['axe'] = self.curve_axe != 'None'
            if self.curve_axe == '1 mm':
                if self.data_type == 'Raw':
                    y = info['y_offset'] + 100 - info['norm_min']
                elif self.data_type == 'Log':
                    y = info['y_offset'] + np.log(100) - info['norm_min']
                else:
                    logger.warning('Curve axe 1 mm not available in Detrend mode, set to 0 mm')
                    y = info['y_offset'] - info['norm_min']
            elif self.curve_axe == '0 mm':
                y = info['y_offset'] - info['norm_min']
            elif self.curve_axe == 'Mean':
                y = info['y_offset'] + np.nanmean(row[DATA_VALUES])
            if info['axe']:
                info[f'axe_{HEARTWOOD}'] = init_ColumnDataSource()
                info[f'axe_{HEARTWOOD}'].data['y'] = [y]
                info[f'axe_{SAPWOOD}'] = init_ColumnDataSource()
                info[f'axe_{SAPWOOD}'].data['y'] = [y]
        
        def get_curve_axe_value(row, info):
            if info['axe']:
                info[f'axe_{HEARTWOOD}'].data['x0'] = [info[HEARTWOOD].data['x'][0]]
                info[f'axe_{HEARTWOOD}'].data['x1'] = [info[HEARTWOOD].data['x'][-1]]
                info[f'axe_{SAPWOOD}'].data['x0'] = [info[SAPWOOD].data['x'][0]]
                info[f'axe_{SAPWOOD}'].data['x1'] = [info[SAPWOOD].data['x'][-1]]
            
        def get_signatures(row, info):
            if row[CATEGORY] != MEAN:
                return
            sig = row[DATA_SIGNATURES]
            values = row[DATA_VALUES]
            if sig is None:
                return
            if self.signature_75:
                mask = (sig >= 0.75) #& (sig < 0.9)
                mask = np.concatenate([(mask[:-1] | mask[1:]), mask[-1:]])
                v = copy.deepcopy(values)
                v[~mask] = np.nan
                info['signature_75'] = init_ColumnDataSource()
                info['signature_75'].data['x'] = np.arange(0, len(sig)) + info['x_offset']
                info['signature_75'].data['w'] = v
                info['signature_75'].data['y'] = info['signature_75'].data['w'] + info['y_offset']
            if self.signature_90:
                mask = (sig >= 0.9)
                mask = np.concatenate([(mask[:-1] | mask[1:]), mask[-1:]])
                v = copy.deepcopy(values)
                v[~mask] = np.nan
                info['signature_90'] = init_ColumnDataSource()
                info['signature_90'].data['x'] = np.arange(0, len(sig)) + info['x_offset']
                info['signature_90'].data['w'] = v
                info['signature_90'].data['y'] = info['signature_90'].data['w'] + info['y_offset']
        
        def get_y(info, dy):
            if self.anatomy == 'Axe' and info['axe']:
                return info[f'axe_{HEARTWOOD}'].data['y'][0]
            return dy + info['y_offset']
        
        draw = {}
        dgroup = {}
        data = data.loc[data[CATEGORY].isin([MEAN, MEASURE]),:]
        if self.x_offset_mode != 'None':
            if data[self.x_offset_mode].isna().any():
                logger.error(f"NA value(s) in {self.x_offset_mode} column, can't draw")
                self.draw_data = None
                return
                    
        cum_y_offset = 0
        group_by = self.group_by if self.group_by is not None else [0]*len(data)
        for group, df in data.groupby(group_by):
            grp = {}
            grp['name'] = group
            grp['y0'] = cum_y_offset
            for i, (_, row) in enumerate(reversed(list(df.iterrows()))):  
                #perror(f'\t row : {row[KEYCODE]}') 
                info = {}
                info['data'] = 'wood'
                info[CATEGORY] = row[CATEGORY]
                row[RAW] = row[DATA_VALUES]

                info['norm_min'] = 0
                if self.y_height == '[min, max]':
                    info['norm_min'] = np.nanmin(row[DATA_VALUES])
                    row[DATA_VALUES] -= info['norm_min']                 
                if self.draw_type == 'Timeline':
                    row[DATA_VALUES] = np.array([100] * row[DATA_LENGTH])
                info['i'] = i
                info[KEYCODE] = row[KEYCODE]
                info['x_offset'] = get_x_offset(row)
                info[DATA_LENGTH] = row[DATA_LENGTH]
                h, next_cum_y_offset = get_y_offset(row, cum_y_offset)
                #print(i, row[KEYCODE], cum_y_offset, h, next_cum_y_offset)
                info['y_offset'] = cum_y_offset
                get_curve_axe(row, info)
                get_values(row, info)
                get_bark(row, info)
                get_missing_ring(row, info)
                get_min_max(row, info)
                get_border(row, info)
                get_text(row, i, info)
                get_curve_axe_value(row, info)
                get_signatures(row, info)
                info['legend_group'] = group
                
                draw[info[KEYCODE]] = info
                cum_y_offset = next_cum_y_offset
            grp['y1'] = cum_y_offset
            dgroup[group] = grp

        self.draw_data = draw
        self.draw_group = dgroup 
        #print('prepare_data end')


    def clear(self):
        self.figure_pane.object =  figure(margin=(5), title='', toolbar_location="left", height=self.height, width=self.width,
                tools="pan,wheel_zoom,box_zoom,reset,hover,save,crosshair,tap,box_edit,freehand_draw")


    #@param.depends('y_delta', 'y_height', 'x_offset_mode', 'anatomy', 
    #               'y_offset_mode', 'draw_type', 'cambium_estimation', 'group_by', 
    #               'legend', 'legend_location', 'show_dates', 'show_estimated_dates', 
    #               'show_cambium_season', 'curve_axe', watch=True)
    def prepare_and_plot_inner(self):
        self.prepare_and_plot(data=self.data)
    
    def prepare_and_plot(self, data=None):
        try:
            #self._layout.loading = True
            if (data is None) or (len(data) == 0):
                self.clear()
                #self._layout.loading = False
                return
            self.prepare_data(data) 
        except Exception as inst:
            logger.error(f'plot : {inst}', exc_info=True)
        #finally:
            #self._layout.loading = False
        
        self.plot()
        #print('end plot')
    
    def on_x_range_step(self, fig=None):
        if fig is None:
            fig = self.figure_pane.object
        if (fig is not None) and pd.notna(fig.x_range.start):
            x_min = fig.x_range.start + self.axis_marge 
            x_max = fig.x_range.end - self.axis_marge + self.x_range_step
            fig.xaxis[0].ticker = FixedTicker(ticks= np.arange(int(x_min), int(x_max), self.x_range_step))
            label = self.x_offset_mode if self.x_offset_mode != 'None' else f'{OFFSET}'
            fig.xaxis[0].axis_label = label
        
    #@param.depends('figure_title', 'figure_title_font_size', watch=True)
    def on_figure_title(self, fig=None):
        if fig is None:
            fig = self.figure_pane.object
        if fig is not None:
            fig.title.text = self.figure_title
            fig.title.text_font_size = str(self.figure_title_font_size) + 'px'
    
    #@param.depends('axis_font_size', watch=True)
    def on_axis_font_size(self, fig=None):
        if fig is None:
            fig = self.figure_pane.object
        if fig is not None:
            fig.yaxis.major_label_text_font_size = f'{self.axis_font_size}px'
            fig.xaxis.major_label_text_font_size = f'{self.axis_font_size}px'
    
    #@param.depends('legend_font_size', watch=True)
    def on_legend_font_size(self, fig=None):
        if fig is None:
            fig = self.figure_pane.object
        if fig is not None:
            fig.legend.label_text_font_size = f'{self.legend_font_size}px'
            #print(fig.legend.label_text_font_size)

    def on_legend_location(self, fig=None):
        if fig is None:
            fig = self.figure_pane.object
        if fig is not None:
            if self.legend_location not in ['None', 'Axe Y', 'Curve begin', 'Curve end']:
                #print(self.legend_location, str(self.legend_location))
                fig.legend.location = self.legend_location
                
    #@param.depends('grid_line_visible', watch=True)
    #def on_grid_line_color(self):
    #    if fig is not None:
    #        fig.ygrid.grid_line_color = fig.xgrid.grid_line_color if self.grid_line_visible else None

    def get_legend(self, i, keycode):
        if self.legend == 'Number+' + KEYCODE: 
            return f'[{i}] {keycode}'
        elif self.legend == 'Number':
            return str(i)
        elif self.legend == KEYCODE:
            return keycode
        return ''

    def on_group(self, fig=None):
        if fig is None:
            fig = self.figure_pane.object
        if self.group_by is None:
            return
        
        x0 = fig.x_range.start
        x1 = fig.x_range.end
        fs = str(self.legend_font_size)+'px'
        if len(self.draw_group) <= 10:
            colors = Category10[max(len(self.draw_group), 3)]
        else:
            colors = Category20[len(self.draw_group)] if len(self.draw_group) <= 20 else Category20[20]
        boxes = []
        for id, (grp, info) in enumerate(self.draw_group.items()):
            color = colors[id % len(colors)]
            boxes.append(BoxAnnotation(top=info['y1'], bottom=info['y0'], left=x0, right=x1,  fill_alpha=0.1, fill_color=color, 
                                       border_radius = 5, line_width=1, line_color=color, editable = True, resizable='all', movable='both'))
            
            label = Label(x=x0, y=info['y1'], anchor="top_left", text=str(grp), editable = True, 
                  text_color='black', text_font_size=fs, text_font_style='bold italic', x_offset=10, y_offset=0)
            fig.add_layout(label)
            
        fig.renderers.extend(boxes)
        
    def on_legend(self, fig=None):
        if fig is None:
            fig = self.figure_pane.object
        if fig is not None:            
            y_labels = {}
            for i, (keycode, info) in enumerate(self.draw_data.items()):                 
                y_labels[info['y_label']] = self.get_legend(i, keycode)
            #print(y_labels)
            #print('-'*10)
            #print(self.legend_location)
            if self.legend_location == 'None':
                fig.legend.visible = False
                fig.yaxis.visible = False
            elif self.legend_location == 'Axe Y':
                fig.legend.visible = False
                fig.yaxis.visible = True
                fig.yaxis.ticker = list(y_labels.keys())
                fig.yaxis.major_label_overrides = y_labels
                #print(fig.yaxis.major_label_overrides)
            elif self.legend_location.startswith('Curve'):
                fig.legend.visible = False
                fig.yaxis.visible = False
                fig.ygrid.grid_line_color = None
            else:
                fig.legend.visible = True
                fig.yaxis.visible = True
                fig.yaxis.ticker = list(y_labels.keys())
                fig.yaxis.major_label_overrides = y_labels
                #print(fig.yaxis.major_label_overrides)
                fig.legend.location = self.legend_location #"top_left"
                fig.legend.click_policy="mute"
                fig.ygrid.grid_line_color = None
       
    def get_color(self, kind, rank):
        if self.color == self.ANATOMY:
            return self.ANATOMY_COLOR[kind]
        elif self.color == KEYCODE:
            if len(self.draw_data)  <= 10:
                return Category10[10][rank]
            else:
                return Category20[20][rank % 20]
        return 'black'
    
    def get_label_legend(self, i, info):
        return f'{i} - {info[KEYCODE]}'
    
    #@param.depends( 'border', 'x_range_step', 
    #               'line_width_tree', 'line_width_mean', 'circle_radius', 'color', 'axis_marge', 'legend_font_size',
    #               watch=True)
    def plot(self, x_range = None, y_range = None):   
        #logger.debug('plot')
        #self._layout.loading = True
        try:

            # save x_range and y_range values for next plot (usefull for ligt ploter)
            if x_range is not None:
                self.x_range = x_range
            if y_range is not None:
                self.y_range = y_range
                
            #('ploter')
            if self.draw_data is None:
                return
            
            fig = figure(margin=(5), title=self.figure_title, toolbar_location="left", #height=self.height, width=self.width,
                tools="pan,wheel_zoom,box_zoom,reset,hover,save,crosshair,tap,box_edit,freehand_draw", tooltips=[('(date/offset,value)', '(@x, @w)')],
                )
            
            fig.output_backend = "svg"
            radius = self.marker_size
            line_dash = [6, 3]
            
            x = []
            for i, (keycode, info) in enumerate(self.draw_data.items()):
                line_width = self.line_width_tree if info[CATEGORY] == MEASURE else self.line_width_mean
                x.append(info['x_min'])
                x.append(info['x_max'])
                fct = fig.line
                if self.draw_type == 'Step':
                    fct = fig.step
                info['ids'] = []
                fct(x='x', y='y', source=info[HEARTWOOD], line_width=line_width,  color=self.get_color(HEARTWOOD, i), legend_label=self.get_label_legend(i, info))
                fct(x='x', y='y', source=info[SAPWOOD], line_width=line_width,  color=self.get_color(SAPWOOD, i), legend_label=self.get_label_legend(i, info))
                if info['is_cambium_estimated'] and self.cambium_estimation:
                    fig.segment(x0='x0', y0='y0', x1='x1', y1='y1', source=info[CAMBIUM_ESTIMATED], line_width=line_width, color=self.get_color(SAPWOOD, i), legend_label=self.get_label_legend(i, info))
                    fct(x='x', y='y', source=info[CAMBIUM_BOUNDARIES], line_dash=line_dash, line_width=line_width, color=self.get_color(SAPWOOD, i), legend_label=self.get_label_legend(i, info))
                if self.border:
                    fig.quad(top='top', bottom='bottom', left='left', right='right', source=info['border'], line_color='red', alpha=0.05, line_width=1, color='red')
                if MISSING_RING_BEGIN in info:
                    fct(x='x', y='y', source=info[MISSING_RING_BEGIN], line_dash=line_dash, line_width=line_width, color=self.get_color(HEARTWOOD, i), legend_label=self.get_label_legend(i, info))
                if MISSING_RING_END in info:
                    c = self.get_color(SAPWOOD, i) if info['is_sapwood'] else self.get_color(HEARTWOOD, i)
                    fct(x='x', y='y', source=info[MISSING_RING_END], line_dash=line_dash, line_width=line_width, color=c, legend_label=self.get_label_legend(i, info)) 
                    
                fig.scatter(x='x', y='y', source=info[PITH], size=radius,marker="circle" ,color=self.get_color(PITH, i), legend_label=self.get_label_legend(i, info))
                fig.scatter(x='x', y='y', source=info[CAMBIUM], size=radius,marker="circle" ,color=self.get_color(SAPWOOD, i), legend_label=self.get_label_legend(i, info))
                fig.scatter(x='x', y='y', source=info[BARK], size=radius, marker="circle",color=self.get_color(BARK, i), legend_label=self.get_label_legend(i, info))
                
                if info['axe']:
                    fig.segment(x0='x0', y0='y', x1='x1', y1='y', source=info[f'axe_{HEARTWOOD}'], line_width=1, color=self.ANATOMY_COLOR[HEARTWOOD])
                    fig.segment(x0='x0', y0='y', x1='x1', y1='y', source=info[f'axe_{SAPWOOD}'], line_width=1, color=self.ANATOMY_COLOR[SAPWOOD])
                
                if self.signature_75 and 'signature_75' in info:
                    fct(x='x', y='y', source=info['signature_75'], line_width=line_width*2,  color=self.get_color('signature_75', i), legend_label=self.get_label_legend(i, info))
                    #fig.scatter(x='x', y='y', source=info['signature_75'], size=radius, marker="asterisk", color=self.get_color('signature_75', i), legend_label=self.get_label_legend(i, info))
                if self.signature_90 and 'signature_90' in info:
                    fct(x='x', y='y', source=info['signature_90'], line_width=line_width*2,  color=self.get_color('signature_90', i), legend_label=self.get_label_legend(i, info))
                    #fig.scatter(x='x', y='y', source=info['signature_90'], size=radius, marker="asterisk", color=self.get_color('signature_90', i), legend_label=self.get_label_legend(i, info))
                
                fs = str(self.legend_font_size)+'px'
                fig.text(x='x', y='y', text='text', source=info['text_begin'], x_offset=-5 , y_offset=0, anchor="center_right", text_font_size=fs)
                obj = fig.text(x='x', y='y', text='text', source=info['text_end'], x_offset=5 , y_offset=0, anchor="center_left", text_font_size=fs)
                   
            (x_min, x_max) = (np.min(x), np.max(x)) if self.x_range is None else self.x_range
            fig.x_range = Range1d(start=x_min - self.axis_marge, end=x_max + self.axis_marge)
            if self.y_range is not None:
                fig.y_range = Range1d(self.y_range[0], self.y_range[1])

            fig.legend.visible = False
            
            self.on_x_range_step(fig)
            self.on_legend(fig)
            self.on_group(fig)
            self.on_figure_title(fig)
            self.on_axis_font_size(fig)
            self.on_legend_font_size(fig)
            self.figure_pane.object = fig
            self._update_width()
            self._update_height()
            #perror('plot', self.width, self.height, self.figure_pane.object.width, self.figure_pane.object.height)
        except Exception as inst:
            logger.error(f'plot : {inst}', exc_info=True)
        #finally:
        #    self._layout.loading = False


# import numpy as np
# import pandas as pd
# import matplotlib.pyplot as plt
# from scipy.signal import hann

# def skel_plot(rw_vec, yr_vec=None, sname="", filt_weight=9, dat_out=False, master=False, plot=True):
#     # Validate and handle sname
#     if not sname:
#         sname2 = ""
#     else:
#         sname2 = str(sname)
#         if len(sname2) > 7:
#             raise ValueError("'sname' must be a character string whose width is less than 8")

#     # Remove NaN values from rw_vec and prepare series
#     na_mask = np.isnan(rw_vec)
#     rw_vec2 = rw_vec[~na_mask]
#     n_val = len(rw_vec2)

#     if n_val > 840:
#         raise ValueError("Long series (> 840) must be split into multiple plots")

#     if filt_weight < 3:
#         raise ValueError("'filt.weight' must be greater or equal to 3")

#     if n_val < filt_weight:
#         raise ValueError("'filt.weight' must not be larger than length of input series")

#     # Handle yr_vec
#     if yr_vec is None:
#         yr_vec2 = np.arange(n_val)
#     else:
#         yr_vec2 = yr_vec[~na_mask]

#     # Pad to the nearest decade if needed
#     pad0 = (min(yr_vec2) // 10) * 10
#     if pad0 != min(yr_vec2):
#         pad_length = min(yr_vec2) - pad0
#         pad_data = pd.DataFrame({'rw': [np.nan] * pad_length, 'yr': np.arange(pad0, pad0 + pad_length)})
#         rw_df = pd.concat([pad_data, pd.DataFrame({'rw': rw_vec2, 'yr': yr_vec2})], ignore_index=True)
#     else:
#         rw_df = pd.DataFrame({'rw': rw_vec2, 'yr': yr_vec2})

#     # Detrend using a Hanning window
#     rw_rw = rw_df['rw'].to_numpy()
#     window = hann(filt_weight)
#     rw_dt = np.convolve(rw_rw, window / window.sum(), mode='same')

#     skel = np.full_like(rw_rw, np.nan, dtype=float)

#     # Calculate relative growth
#     temp_diff = np.diff(rw_rw)
#     idx = np.arange(1, len(rw_rw) - 1)
#     skel[idx] = (temp_diff[:-1] + temp_diff[1:]) / (2 * rw_dt[idx])

#     skel[skel > 0] = np.nan  # Keep only negative growth
#     non_na_mask = ~np.isnan(skel)

#     if non_na_mask.any():
#         skel_range = [skel[non_na_mask].min(), skel[non_na_mask].max()]
#         new_range = [10, 1]
#         scale_factor = (new_range[1] - new_range[0]) / (skel_range[1] - skel_range[0])
#         skel = new_range[0] + (skel - skel_range[0]) * scale_factor
#         skel[skel < 3] = np.nan
#         skel = np.ceil(skel)

#     # Plotting if required
#     if plot:
#         plt.figure(figsize=(14, 9))
#         plt.plot(rw_df['yr'], skel, label='Skeleton Plot', color='black')
#         plt.xlabel('Year')
#         plt.ylabel('Skeleton Value')
#         plt.title(f'Skeleton Plot for {sname2}')
#         plt.grid(True)
#         plt.show()

#     if dat_out:
#         return pd.DataFrame({'year': rw_df['yr'], 'skeleton': skel})

# # Example usage:
# # rw_vec = np.array([...])  # Replace with your data
# # yr_vec = np.array([...])  # Replace with your year data if applicable
# # skel_plot(rw_vec, yr_vec, sname="Sample")




"""
    pyodide version of pyDendron
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
 
import os
#import getpass
#import argparse
#import sys

import panel as pn
from pathlib import Path
#import json
#from bokeh.server.contexts import BokehSessionContext
import bokeh
#from datetime import datetime
#import platform
#import shutil

from pyDendron.dataset import Dataset
from pyDendron.gui_panel.sidebar import ParamColumns, ParamDetrend, ParamMean, ParamPackage, ParamColumnPackage
from pyDendron.gui_panel.dataset_selector import DatasetSelector
from pyDendron.gui_panel.dataset_treeview import DatasetTreeview
from pyDendron.gui_panel.dataset_package import DatasetPackage
from pyDendron.gui_panel.dataset_package_builder import DatasetPackageBuilder
from pyDendron.gui_panel.dataset_package_editor import DatasetPackageEditor
from pyDendron.gui_panel.tools_panel import ToolsPanel
from pyDendron.gui_panel.ploter_panel import PloterPanel
from pyDendron.gui_panel.debug_panel import DebugPanel
from pyDendron.gui_panel.crossdating_panel import CrossDatingPanel
from pyDendron.app_logger import logger, perror, catch_bokeh_log, open_mail_client, LOG_FILENAME, __version__


def bt_notification():
    """
    Create a button to clear notifications.
    
    Returns:
        pn.widgets.Button: The button to clear notifications.
    """
    def on_rm_notification(event):
        pn.state.notifications.clear()
        
    rm_notification = pn.widgets.Button(name='Clear notifications', icon="trash", button_type='default', align=('end', 'center'))
    rm_notification.on_click(on_rm_notification)
    return rm_notification

def bt_gps():
    """
    Create a button to open the GPS site.
    
    Returns:
        pn.pane.Markdown: The button to open the GPS site.
    """
    # GPS 
    return pn.pane.Markdown("""<svg xmlns="http://www.w3.org/2000/svg" width="18" height="18" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="1" stroke-linecap="round" stroke-linejoin="round" class="icon icon-tabler icons-tabler-outline icon-tabler-map">
    <path stroke="none" d="M0 0h24v24H0z" fill="none" />
    <path d="M3 7l6 -3l6 3l6 -3v13l-6 3l-6 -3l-6 3v-13" />
    <path d="M9 4v13" />
    <path d="M15 7v13" />
    </svg>
    <a href="https://nominatim.openstreetmap.org" target="_blank">Open GPS site</a>""")

def get_logout():
    return pn.pane.Markdown("""<svg  xmlns="http://www.w3.org/2000/svg"  width="18"  height="18"  viewBox="0 0 24 24"  fill="none"  stroke="currentColor"  stroke-width="1"  stroke-linecap="round"  stroke-linejoin="round"  class="icon icon-tabler icons-tabler-outline icon-tabler-logout">
        <path stroke="none" d="M0 0h24v24H0z" fill="none"/><path d="M14 8v-2a2 2 0 0 0 -2 -2h-7a2 2 0 0 0 -2 2v12a2 2 0 0 0 2 2h7a2 2 0 0 0 2 -2v-2" />
        <path d="M9 12h12l-3 -3" /><path d="M18 15l3 -3" /></svg>
        <a href="/logout" >Logout</a>""")    


def bt_send_log():
    def on_send_log(event):
        
        panel_version = pn.__version__
        bokeh_version = bokeh.__version__
        date_today = datetime.now()
        os_info = platform.system() + ' ' + platform.release()
        user_name = pn.state.user #getpass.getuser()
        #perror(f'User: {user_name}')
        
        agent = pn.state.headers["User-Agent"] if "User-Agent" in pn.state.headers else "Unknown"
        
        info_string = f"""
        pyDendron: {__version__}
        Date: {date_today}
        OS: {os_info}
        User: {user_name}
        Agent: {agent}
        Panel: {panel_version}
        Bokeh: {bokeh_version}"""

        info_string += '\n'+'-'*80 + '\n'
        with open(LOG_FILENAME, 'r') as file:
            info_string += file.read()
            
        open_mail_client(
            to='pyDendron@univ-lemans.fr',
            subject=f'pyDendron error {user_name}',
            body=info_string
        )
    
    bt_log = pn.widgets.Button(name='Send log', icon="send", button_type='default', align=('end', 'center'))
    bt_log.on_click(on_send_log)
    return bt_log

def get_version():
    return pn.pane.Markdown(f"version: {__version__} ")


pn.extension('tabulator', 'filedropper', throttled=True, notifications=True, loading_spinner='dots')
pn.param.ParamMethod.loading_indicator = True
#pn.extension(
#    disconnect_notification="""Server Connected Closed <br /> <button class="btn btn-outline-light" onclick="location.reload();">Click To Reconnect</button> """
#)

#logo = os.path.join(os.path.dirname(__file__), 'data', 'trees.svg')
template = pn.template.FastListTemplate(title='pyDendron', 
                                        #logo=logo,
                                        meta_author='LIUM, Le Mans Université',
                                        meta_keywords='Dendromean',
                                        main_layout=None,
                                        )
template.header.objects = [get_version(), bt_notification(), #bt_send_log(), 
                           bt_gps()]

dataset = Dataset()
filters = ['*.p', '*.json', '*.xlsx']
#perror('dataset_selector', )
dataset_selector = DatasetSelector(dataset, template=template, path='file::/Users/meignier/pyDendron/dataset/meignier', filters=filters, cfg_path='')    
#dataset_selector.wselect.options = ['test', 'toto', 'tata'] 
template.main.append(dataset_selector)

template.servable()




"""
Debug and log class
"""
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
__license__ = "GPL"

import pandas as pd
import panel as pn
import param
import sys
from panel.viewable import Viewer

from pyDendron.dataname import *
from pyDendron.app_logger import logger#, stdout_stream_handler

class DebugPanel(Viewer):
    """ Tools to manage dataset. """
        
    def __init__(self, dataset, parameters, app, **params):
        super().__init__(**params)
        self.dataset = dataset
        self.app = app
        self.parameters = parameters
        self.wlog = None
        self.wcrossdating_log = None
        
        # Logger
        #self.terminal = pn.widgets.Terminal(
        #    "Debugger terminal\n\n",
        #    options={"cursorBlink": True},
        #    sizing_mode='stretch_both'
        #)
        #sys.stdout = self.terminal
        #stdout_stream_handler.setStream(self.terminal)
        
        self._layout = pn.Tabs(
            #('Debug', self.terminal), 
            ('Dataset Log', self._log()),
            ('Crossdating Log', self._crossdating_log()),
            #('id Log', self._id()),
            dynamic=False, margin=0, styles={'font-size': '16px'}, name=self.name)
        
        self._layout.param.watch(self.on_tab_change, ['active'], onlychanged=True)

    def __panel__(self):
        """ Return the panel layout."""
        return self._layout

    def _id(self):
        def get_id(event):
            from bokeh.io import curdoc
            doc = curdoc()
            
            id = int_input.value
            #print(f"Get id {id}")
            #print(doc.get_model_by_id(id))

        int_input = pn.widgets.IntInput(name='Bokeh id', value=0, description='Enter the Bokeh id to get the model.')
        button = pn.widgets.Button(name='Get', button_type='primary', description='Get the model.')
        button.on_click(get_id)
        
        return pn.Column(int_input, button)
        
    def _log(self):
        def get_log(event):
            self.wlog.value = self.dataset.get_log()
        
        button = pn.widgets.Button(name='Load', button_type='primary', description='Load log.')
        button.on_click(get_log)

        self.wlog = pn.widgets.Tabulator(
                            show_index=False,
                            sizing_mode='stretch_width',
                            max_height=800,
                            min_height=400,
                            height_policy='max',
                            )
        return pn.Column(
            button,
            self.wlog
        )
    
    def _crossdating_log(self):
        def get_log(event):
            self.wcrossdating_log.value = self.dataset.get_crossdating_log()
        
        button = pn.widgets.Button(name='Load', button_type='primary', description='Load crossdating log.')
        button.on_click(get_log)

        self.wcrossdating_log = pn.widgets.Tabulator(
                            show_index=False,
                            sizing_mode='stretch_width',
                            max_height=800,
                            min_height=400,
                            height_policy='max',
                            )
        return pn.Column(
            button,
            self.wcrossdating_log
        )

    def on_tab_change(self, event):
        """ Handle tab change event. """
        if self._layout.active == 0:
            self.wlog.value = self.dataset.get_log()
        elif self._layout.active == 1:
            self.wcrossdating_log.value = self.dataset.get_crossdating_log()



"""
Pith / Cambium estimation
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"

import numpy as np
import pandas as pd

def cambium_estimation(param, cambium, bark, sapwood, values):  
    if pd.isna(cambium) or pd.isna(bark): # no cambium or bark
        return pd.NA, pd.NA, pd.NA
    if cambium or bark: # cambium or bark persent
        return pd.NA, pd.NA, pd.NA
    elif pd.isna(sapwood) or ((sapwood > 0) and (sapwood >= len(values))) : # no sapwood
        return pd.NA, pd.NA, pd.NA
    elif param.cambium_estimation_method == 'log-log':
        return cambium_estimation_log_log(cambium, bark, sapwood, values)
    elif param.cambium_estimation_method == 'Lambert':
        return cambium_estimation_lambert(param.lambert_parameters, cambium, bark, sapwood, values)
    else:
        return pd.NA, pd.NA, pd.NA

def cambium_estimation_log_log(cambium, bark, sapwood, values):    
    nb = len(values)
    i = max(0, sapwood - 9)
    j = min(sapwood + 1, nb)
    rw10 = np.nanmean(values[i:j]) / 100
    x = 2.8102081 - 0.5331451 * np.log(rw10)
    estimation = int(np.round(np.exp(x))) + sapwood
    upper_bound = int(np.round(np.exp(x + 0.6087837))) + sapwood
    lower_bound = int(np.round(np.exp(x - 0.6087837))) + sapwood
    #print(f'sapwood={spawood}, estimation={estimation}, upper_bound={upper_bound}, lower_bound={lower_bound}')
    return lower_bound , estimation, upper_bound

def cambium_estimation_lambert(estimator, cambium, bark, sapwood, values):    
    lower_bound = estimator[0] + sapwood
    estimation = (estimator[0] + estimator[1]) // 2 + sapwood
    upper_bound = estimator[1] + sapwood
    return lower_bound , estimation, upper_bound




""" I/O for Besançon format """

__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
__license__ = "GPL"

from pathlib import Path
import pandas as pd
import numpy as np
import io

from pyDendron.dataname import *
from pyDendron.alien.io import IO
from pyDendron.dataset import Dataset
from pyDendron.app_logger import logger, perror

class IOBesancon(IO):
    SPECIES_MAPPING = {
        'Q': 'QU',
        'QU': 'QU',
        'CHENE': 'QU'
    }
    file_extension = ['.txt', '.bes', '.ne']
    
    def read_properties(self, buffer, meta):
        def species(word):
            return self.SPECIES_MAPPING.get(word, word)
            #if word in ['Q', 'QU', 'CHENE']:
            #    return 'QU';
            #else:
            #    return word
    
        def pith(word):
            return True
        
        def cambium(word): 
            return True if word == 'OUI' else False
        
        def integer_(word):
            return int(word)

        prop = {
            'ESP': (species, SPECIES), 
            #'MOE': (pith, PITH), 
            #'CAM': (cambium, CAMBIUM), 
            'POS': (integer_, OFFSET), 
            'ORI': (integer_, DATE_BEGIN), 
            'TER': (integer_, DATE_END), 
            #'MXTER': (integer_, DATE_END_ESTIMATED), 
            'AUB': (integer_, SAPWOOD)
        }
        meta[OFFSET] = pd.NA
        meta[CATEGORY] = MEASURE
        
        words = buffer.upper().split() + ['UNK']
        for i, word in enumerate(words):
            for key, (fct, k) in prop.items():
                if word.startswith(key):
                    meta[k] = fct(words[i+1])
                    key_found = True

        if meta[DATE_BEGIN] < 0: meta[DATE_BEGIN] += 1
        if meta[DATE_END] > 0: meta[DATE_END] += 1
                    
    def read_values(self, buffer, meta):
        words = buffer.upper().replace(';', '').split()   
        values = [int(word) if word not in [',', '!', '?', '*', '/'] else np.nan for word in words]
        #values = [] 
        #for word in words:
        #    if word not in[',', '!', '?', '*', '/', ';']:
        #        values.append(int(word))
        #    elif word == ',':
        #        values.append(np.nan)
        meta[DATA_VALUES] = np.array(values, dtype='float')
        meta[DATA_LENGTH] = len(np.array(values))
        
    
    def read_comps(self, buffer, meta, id_comps):
        lines = buffer.upper().split('\n')   
        for line in lines:
            c, keycode, offset = line.split()
            if keycode in id_comps:
                meta[id_comps[keycode]] = offset
            else:
                raise ValueError(f'{keycode} sequence is not in the file')
    
    def read_sequences(self, id_parent, lines):
        meta = {}
        state = 'start'
        buffer = ''
        id_comps = {}
        for line in lines:
            if line[0] == '.': # new sequences
                if state == 'value':
                    self.read_values(buffer, meta)
                elif state == 'comp':
                    self.read_comps(buffer, meta, id_comps)             
                if state != 'start':
                    self.components.append({ID_PARENT: id_parent, ID_CHILD: meta[ID], OFFSET: meta[OFFSET]})
                    del meta[OFFSET]
                    self.sequences.append(meta)

                meta = {ID: self.next_id()}
                meta[KEYCODE] = line.split()[1]
                id_comps[meta[KEYCODE]] = meta[ID]
                state = 'seq'
                buffer = ''
            elif line[0] == 'V': # new values
                if state != 'seq':
                    raise ValueError('inconsistent state: {state}')
                self.read_properties(buffer, meta)
                state = 'value'          
                ring_type = line.split()[1].upper()
                if not ring_type.startswith('NAT'):
                    logger.warning(f'{DATA_TYPE} is not raw')
                    meta[DATA_TYPE] = ring_type
                else:
                    meta[DATA_TYPE] = RAW
                buffer = ''
            elif line[0] == 'C': # new composent
                if state != 'value':
                    raise ValueError('inconsistent state: {state}')
                state = 'comp'
                self.read_values(buffer, meta)
                buffer = ''
            elif line[0] == ':': # end
                if state not in  ['value', 'comp']:
                    raise ValueError('inconsistent state: {state}')
                state = 'end'                
                break
            else:
                buffer += line
                
        if state != 'end':
            if state == 'value':
                self.read_values(buffer, meta)
            elif state == 'comp':
                self.read_comps(buffer, meta, id_comps)             
            self.components.append({ID_PARENT: id_parent, ID_CHILD: meta[ID], OFFSET:meta[OFFSET]})
            del meta[OFFSET]
            self.sequences.append(meta)
            

    def write_file(self, data: pd.DataFrame, means, filename: str):
        with open(Path(filename), mode='w', encoding=self.encoding, errors='ignore') as fd:
            for i, row in data.iterrows():
                if row[CATEGORY] in [MEASURE, MEAN]:
                    fd.write(f'. {row[KEYCODE].replace(" ", "_")}\n')
                    if pd.notna(row[DATA_LENGTH]):
                        fd.write(f'\tLON {row[DATA_LENGTH]}\n')
                    if pd.notna(row[OFFSET]):
                        fd.write(f'\tPOS {row[OFFSET]}\n')
                    if pd.notna(row[DATE_BEGIN]):
                        fd.write(f'\tORI {int(row[DATE_BEGIN])}\n')
                    if pd.notna(row[DATE_END]):
                        fd.write(f'\tTER {int(row[DATE_END])}\n')
                    
                    if row[CATEGORY] == MEASURE:
                        if row[SPECIES] != '':
                            fd.write(f'\tESP {row[SPECIES].replace(" ", "_")}\n')
                        #if pd.notna(row[DATE_END_ESTIMATED]):
                        #    fd.write(f'\tMXTER {int(row[DATE_END_ESTIMATED])}\n')
                        if pd.notna(row[SAPWOOD]):
                            fd.write(f'\tAUB {int(row[SAPWOOD])}\n')
                        if pd.notna(row[PITH]) and row[PITH]:
                            fd.write('\tMOE oui\n')
                        if pd.notna(row[CAMBIUM]) and row[CAMBIUM]:
                            fd.write('\tCAM oui\n')
                    
                    if row[DATA_VALUES] is not None:
                        fd.write(f'VALeurs {row[DATA_TYPE]}\n')
                        tmp = np.round(np.nan_to_num(row[DATA_VALUES], nan=-9999), 0).astype(int).tolist()
                        
                        #tmp = np.nan_to_num(row[DATA_VALUES], nan=-9999).astype(int)
                        tmp = np.array_split(tmp, len(tmp) // 10 + 1)
                        for i, line in enumerate(tmp):
                            lst = [',' if x == -9999 else str(x) for x in line.tolist() ]
                            s = ' '.join(lst)
                            s += ' ;\n' if i + 1 == len(tmp) else '\n'
                            fd.write(s)

                    if row[CATEGORY] == MEAN:
                        fd.write(f'Composante {row[KEYCODE].replace(" ", "_")}\n')
                        for keycode, offset in means[row[ID_CHILD]]:
                            fd.write(f'C    {keycode.replace(" ", "_")}    offset\n')
                    fd.write('\n') 
                    
            

    
    
    
    



"""
Dataset selector
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"

import param
import panel as pn
from panel.viewable import Viewer
import json

from pathlib import Path

from pyDendron.app_logger import logger, perror
from pyDendron.dataname import *
from pyDendron.dataset import Dataset
from pyDendron.gui_panel.my_viewer import MyViewer
from pyDendron.gui_panel.dialog import MessageBox

class DatasetSelector(MyViewer):
    """
    Show dataset tree as list of panel 
    """
    path = param.Foldername('./', doc='path of the data')
    filters = param.List(['*.p', '*.xlsx', '*.json'], doc='glob filter')
    options = param.Dict({}, doc='options')
    save_auto =  param.Boolean(True, doc='Automatically save the dataset on every major change or every 5 minutes.')
    save_cfg =  param.Boolean(True, doc='Save dataset and configuration when disconnected.')
    backup =  param.Boolean(True, doc='Create a backup of the dataset when loading a new dataset.')
    
    def __init__(self, dataset, template, cfg_path, **param):
        super(DatasetSelector, self).__init__(cfg_path, **param)
        #self.cfg_file = cfg_path / Path('pyDendron.selector.cfg.json')
        self.template = template
        self.dataset = dataset

        self.wselect = pn.widgets.Select(name='Selection', options=self.get_options(), description='Dataset to load.')
        
        self.bt_load = pn.widgets.Button(name='Load', icon='loader', button_type='primary', description='Load the selected dataset.')
        self.bt_load.on_click(self.on_load)
        self.bt_refresh = pn.widgets.Button(name='Refresh', icon='refresh', button_type='primary', description='Refresh the dataset list.')
        self.bt_refresh.on_click(self.on_refresh)
        self.bt_save = pn.widgets.Button(name='Save', icon='file-download', button_type='primary', description='Save the loaded dataset.')        
        self.bt_save.on_click(self.on_save)
        
        self._layout = pn.Column(self.wselect, 
                                 pn.Row(self.bt_load, self.bt_refresh, self.bt_save), 
                                 pn.Row(self.param.save_auto, self.param.backup, self.param.save_cfg)
                                 )

        self.param.watch(self.on_save_auto,  ['save_auto'], onlychanged=True)
        
    def __panel__(self):
        return self._layout

    def get_sidebar(self):
            """
            Returns a Card widget that contains the sidebar parameters.

            Returns:
                pn.Card: The Card widget representing the sidebar.
            """
            box = pn.Card(
                self._layout, 
                margin=(5, 0), 
                sizing_mode='stretch_width',
                hide_header=True, 
                title='Dataset')
            return box

    def get_options(self):
        """
        Retrieve the available options for the dataset selector.

        Returns:
            dict: A dictionary containing the options for the dataset selector.
                    The keys are formatted as emoji + file name, and the values are the corresponding file paths.
        """
        options = {}
        #perror(f'Filters: {self.filters}')
        for flt in self.filters:
            for file in Path(self.path).glob(flt):
                options[f'\U0001F4E6 {str(file.name)}'] = file
        return options

    def on_load(self, event):
        """
        Event handler for the load button.

        Args:
            event: The event object representing the button click event.

        Raises:
            Exception: If an error occurs during the loading process.

        """
        try:
            self._layout.loading = True
            file = self.wselect.value
            self.dataset.load(file)
            if self.backup:
                self.dataset.backup(file)
        except Exception as inst:
            logger.error(f'DatasetSelector, on_load: {inst}', exc_info=True)
        finally:
            self._layout.loading = False
        
    def on_refresh(self, event):
        """
        Event handler for the refresh button.

        Parameters:
            event : The event object.

        Returns:
            None
        """
        self.wselect.options = self.get_options()

    def on_save_auto(self, event):
        """
        Callback method triggered when the 'save_auto' event is fired.

        Args:
            event: The event object containing information about the event.

        Returns:
            None
        """
        self.dataset.save_auto = event.new
        
    def on_save(self, event):
        """
        Saves the dataset.

        This method is called when the user clicks the save button. 

        Parameters:
        - event: The event object representing the save button click.

        Returns:
        None
        """
        try:
            self._layout.loading = True
            self.dataset.dump()
        except Exception as inst:
            logger.error(f'on_save: {inst}')
        finally:
            self._layout.loading = False

    @property  
    def _values(self):
        return self.wselect.value        






 "Data Validation"
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
__license__ = "GPL"




import pandas as pd
import numpy as np

from pathlib import Path

from pyDendron.app_logger import logger, perror
from pyDendron.sylphe import Sylphe
from pyDendron.dataset import Dataset
from pyDendron.trash.indices import Indices
from pyDendron.healthiness.aligment import Aligment


path=Path('./dataset')

dataset= Dataset.load(path)

keycodes = dataset._sequences.loc[dataset._sequences['type'] == dn.MEASURE, dn.KEYCODE]
ind_seq = dataset.indices.join(keycodes, on=dn.ID, how='inner', lsuffix='.seq', rsuffix='.ind')
ind_seq = Indices(ind_seq.reset_index())

data = Aligment.ndiff_indices(ind_seq)

with open(path / Path('aligments.txt'), 'w') as fic:
    for sublist in data:
        line = ' '.join(map(str, sublist))
        fic.write(line + '\n')




"""
    Adaptation of Viewer class
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
 
import json
import param
import panel as pn
from panel.viewable import Viewer
from pathlib import Path
from pyDendron.app_logger import logger, perror, check_version, __version__

class MyViewer(Viewer):
    def __init__(self, cfg_path, **params):
        super().__init__(**params)
        self.cfg_path = cfg_path
        self.cfg_file = cfg_path / Path(f'{self.__class__.__name__}.cfg.json')
        self.load_cfg()
    
    def dump_cfg(self):
        with open(self.cfg_file, 'w') as fd:
            data = {
                'version': __version__,
                'param' : self.param.serialize_parameters(),
            }
            json.dump(data, fd)

    def load_cfg(self):
        try:
            #print(f'{self.__class__.__name__} load_cfg')
            if check_version(self.cfg_file):
                with open(self.cfg_file, 'r') as fd:
                    data = json.load(fd)
                    for key, value in json.loads(data['param']).items():
                        if key in self.param.params().keys():
                            if key != 'name':
                                self.param.set_param(key, value)
        except Exception as inst:
            logger.warning(f'ignore cfg {self.__class__.__name__}.')
            logger.error(f'load_cfg: {inst}')    



"""
dplpy tools
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"

import dplpy as dpl
import numpy as np
import pandas as pd
from pyDendron.dataset import Dataset
from pyDendron.trash.indices import Indices

def pyDendron2dlp(dataset, key=Indices.RAW):
    ind = dataset.get_indices_addones().xs(key, level='key', drop_level=True)
    date_min = int(ind[dn.DATE_BEGIN].min())
    date_max = int(ind[dn.DATE_END].max())
    data = pd.DataFrame(columns=ind[dn.KEYCODE], index=range(date_min, date_max))
    vect = np.full(date_max - date_min, np.nan)
    for id, row in ind.iterrows():
        dec = int(row[dn.DATE_BEGIN]) - date_min
        vect = np.full(date_max - date_min, np.nan)
        if row['values'].ndim == 1:
            vect[dec:dec+row['count']] = row['values']
        else:
            vect[dec:dec+row['count']] = row['values'][0]
        data[row[dn.KEYCODE]] = vect
    
    return data

def dpl_chron(dataset, biweight=False, prewhiten=False, plot=False):
    data = pyDendron2dlp(dataset)
    return dpl.chron(data, biweight=biweight, prewhiten=prewhiten, plot=plot)
    